'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

'''
This module defines classes and functions to handle the VCalendar
format.
''' # pylint: disable=W0105

from taskcoachlib.domain.base import Object
from taskcoachlib.domain import date
from taskcoachlib.i18n import _

import time, calendar, datetime

#{ Utility functions

def parseDateTime(fulldate):
    ''' Parses a datetime as seen in iCalendar files into a 
    L{taskcoachlib.domain.date.DateTime} object. '''

    try:
        dt, tm = fulldate.split('T')
        year, month, day = int(dt[:4]), int(dt[4:6]), int(dt[6:8])
        hour, minute, second = int(tm[:2]), int(tm[2:4]), int(tm[4:6])

        if fulldate.endswith('Z'):
            # GMT. Convert this to local time.
            localTime = time.localtime(calendar.timegm((year, month, day, hour, minute, second, 0, 0, -1)))
            year, month, day, hour, minute, second = localTime[:6]
    except Exception, e:
        raise ValueError('Malformed date: %s (%s)' % (fulldate, str(e)))

    return date.DateTime(year, month, day, hour, minute, second)

def fmtDateTime(dt):
    ''' Formats a L{taskcoachlib.domain.date.DateTime} object to a string
    suitable for inclusion in an iCalendar file. '''
    dt = dt.utcfromtimestamp(time.mktime(dt.timetuple()))
    return '%04d%02d%02dT%02d%02d%02dZ' % (dt.year, dt.month, dt.day,
                                           dt.hour, dt.minute, dt.second)

def quoteString(s):
    ''' The 'quoted-printable' codec doesn't encode \n, but tries to
    fold lines with \n instead of CRLF and generally does strange
    things that ScheduleWorld does not understand (me neither, to an
    extent). Same thing with \r. This function works around this. '''

    s = s.encode('UTF-8').encode('quoted-printable')
    s = s.replace('=\r', '')
    s = s.replace('=\n', '')
    s = s.replace('\r', '=0D')
    s = s.replace('\n', '=0A')
    return s

#}

#{ Parsing iCalendar files

class VCalendarParser(object):
    ''' Base parser class for iCalendar files. This uses the State
    pattern (in its Python incarnation, replacing the class of an
    object at runtime) in order to parse different objects in the
    VCALENDAR. Other states are

     - VTodoParser: parses VTODO objects.

    @ivar kwargs: While parsing, the keyword arguments for the
        domain object creation for the current (parsed) object.
    @ivar tasks: A list of dictionaries suitable to use as
        keyword arguments for task creation, representing all
        VTODO object in the parsed file. ''' # pylint: disable=W0511

    def __init__(self, *args, **kwargs):
        super(VCalendarParser, self).__init__(*args, **kwargs)
        self.stateMap = { 'VCALENDAR': VCalendarParser,
                          'VTODO':     VTodoParser,
                          'VNOTE':     VNoteParser }
        self.tasks = []
        self.notes = []
        self.init()

    def init(self):
        ''' Called after a state change. '''
        self.kwargs = {} # pylint: disable=W0201

    def setState(self, state):
        ''' Sets the state (class) of the parser object. '''
        self.__class__ = state
        self.init()

    def parse(self, lines):
        ''' Actually parses the file.
        @param lines: A list of lines. '''

        # TODO: avoid using indexes here, just iterate. This way the
        # method can accept a file object as argument.

        currentLine = lines[0]

        for line in lines[1:]:
            if line.startswith(' ') or line.startswith('\t'):
                currentLine += line[1:]
            else:
                if self.handleLine(currentLine):
                    return
                currentLine = line

        self.handleLine(currentLine)

    def handleLine(self, line):
        ''' Called by L{parse} for each line to parse. L{parse} is
        supposed to have handled the unfolding. '''

        if line.startswith('BEGIN:'):
            try:
                self.setState(self.stateMap[line[6:]])
            except KeyError:
                raise TypeError, 'Unrecognized vcal type: %s' % line[6:]
        elif line.startswith('END:'):
            if line[4:] == 'VCALENDAR':
                return True
            else:
                self.onFinish()
                self.setState(VCalendarParser)
        else:
            try:
                idx = line.index(':')
            except ValueError:
                raise RuntimeError, 'Malformed vcal line: %s' % line

            details, value = line[:idx].split(';'), line[idx + 1:]
            name, specs = details[0], details[1:]
            specs = dict([tuple(v.split('=')) for v in specs])

            if specs.has_key('ENCODING'):
                value = value.decode(specs['ENCODING'].lower())
            if specs.has_key('CHARSET'):
                value = value.decode(specs['CHARSET'].lower())
            else:
                # Some  servers only  specify CHARSET  when  there are
                # non-ASCII characters :)
                # More, Horde encodes in UTF-8 without specifying the charset.
                value = value.decode('UTF-8')

            # If  an item  name ends  with  'TMPL', it's  part of  the
            # template system and has to be eval()ed.

            if name.endswith('TMPL'):
                name = name[:-4]
                context = dict()
                context.update(datetime.__dict__)
                context.update(date.__dict__)
                context['_'] = _
                value = eval(value, context)
                if isinstance(value, datetime.datetime):
                    value = fmtDateTime(value)

            self.acceptItem(name, value)

        return False

    def onFinish(self):
        ''' This method is called when the current object ends. '''
        raise NotImplementedError

    def acceptItem(self, name, value):
        ''' Called on each new 'item', i.e. key/value pair. Default
        behaviour is to store the pair in the 'kwargs' instance
        variable (which is emptied in L{init}). '''
        if name in ('CREATED', 'DCREATED'):
            self.kwargs['creationDateTime'] = parseDateTime(value)
        elif name == 'LAST-MODIFIED':
            self.kwargs['modificationDateTime'] = parseDateTime(value)
        elif name == 'SUMMARY':
            self.kwargs['subject'] = value
        elif name == 'CATEGORIES':
            # Horde escapes the comma, even though it's an actual separator.
            # I didn't found any way to include an actual comma in a Horde
            # "tag".
            self.kwargs['categories'] = map(lambda x: x.rstrip('\\').lstrip(), value.split(','))
        elif name == 'DESCRIPTION':
            self.kwargs['description'] = value.replace('\\n', '\n')
        else:
            self.kwargs[name.lower()] = value


class VTodoParser(VCalendarParser):
    ''' This is the state responsible for parsing VTODO objects. ''' # pylint: disable=W0511

    def onFinish(self):
        if not self.kwargs.has_key('plannedStartDateTime'):
            # This means no planned start date, but the task constructor will
            # take today by default, so force.
            self.kwargs['plannedStartDateTime'] = date.DateTime()

        if self.kwargs.has_key('vcardStatus'):
            if self.kwargs['vcardStatus'] == 'COMPLETED' and \
                   not self.kwargs.has_key('completionDateTime'):
                # Some servers only give the status, and not the date (SW)
                if self.kwargs.has_key('last-modified'):
                    self.kwargs['completionDateTime'] = parseDateTime(self.kwargs['last-modified'])
                else:
                    self.kwargs['completionDateTime'] = date.Now()

        self.kwargs['status'] = Object.STATUS_NONE
        self.tasks.append(self.kwargs)

    def acceptItem(self, name, value):
        if name == 'DTSTART':
            self.kwargs['plannedStartDateTime'] = parseDateTime(value)
        elif name == 'DUE':
            self.kwargs['dueDateTime'] = parseDateTime(value)
        elif name == 'COMPLETED':
            self.kwargs['completionDateTime'] = parseDateTime(value)
        elif name == 'PERCENT-COMPLETE':
            self.kwargs['percentageComplete'] = int(value)
        elif name == 'UID':
            self.kwargs['id'] = value.decode('UTF-8')
        elif name == 'PRIORITY':
            # Okay. Seems that in vcal,  the priority ranges from 1 to
            # 3, but what it means depends on the other client...

            self.kwargs['priority'] = int(value) - 1
        elif name == 'STATUS':
            self.kwargs['vcardStatus'] = value
        else:
            super(VTodoParser, self).acceptItem(name, value)


class VNoteParser(VCalendarParser):
    '''Parse VNote objects.'''

    def onFinish(self):
        # Summary is not mandatory.
        if not self.kwargs.has_key('subject'):
            if 'description' in self.kwargs:
                self.kwargs['subject'] = self.kwargs['description'].split('\n')[0]
            else:
                self.kwargs['subject'] = ''
        self.kwargs['status'] = Object.STATUS_NONE
        self.notes.append(self.kwargs)

    def acceptItem(self, name, value):
        if name == 'X-IRMC-LUID':
            self.kwargs['id'] = value.decode('UTF-8')
        elif name == 'BODY':
            self.kwargs['description'] = value.replace('\\n', '\n')
        elif name == 'CLASS':
            pass
        else:
            super(VNoteParser, self).acceptItem(name, value)

#}

#==============================================================================
#{ Generating iCalendar files.

def VCalFromTask(task, encoding=True, doFold=True):
    ''' This function returns a string representing the task in
        iCalendar format. '''

    encoding = ';ENCODING=QUOTED-PRINTABLE;CHARSET=UTF-8' if encoding else ''
    quote = quoteString if encoding else lambda s: s

    components = []
    components.append('BEGIN:VTODO') # pylint: disable=W0511
    components.append('UID:%s' % task.id().encode('UTF-8'))
    
    if task.creationDateTime() > date.DateTime.min:
        components.append('CREATED:%s' % fmtDateTime(task.creationDateTime()))
        
    if task.modificationDateTime() > date.DateTime.min:
        components.append('LAST-MODIFIED:%s' % fmtDateTime(task.modificationDateTime()))

    if task.plannedStartDateTime() != date.DateTime():
        components.append('DTSTART:%s' % fmtDateTime(task.plannedStartDateTime()))

    if task.dueDateTime() != date.DateTime():
        components.append('DUE:%s' % fmtDateTime(task.dueDateTime()))

    if task.completionDateTime() != date.DateTime():
        components.append('COMPLETED:%s' % fmtDateTime(task.completionDateTime()))

    if task.categories(recursive=True, upwards=True):
        categories = ','.join([quote(unicode(c)) for c in task.categories(recursive=True, upwards=True)])
        components.append('CATEGORIES%s:%s' % (encoding, categories))

    if task.completed():
        components.append('STATUS:COMPLETED')
    elif task.active():
        components.append('STATUS:NEEDS-ACTION')
    else:
        components.append('STATUS:CANCELLED') # Hum...

    components.append('DESCRIPTION%s:%s' % (encoding, quote(task.description())))
    components.append('PRIORITY:%d' % min(3, task.priority() + 1))
    components.append('PERCENT-COMPLETE:%d' % task.percentageComplete())
    components.append('SUMMARY%s:%s' % (encoding, quote(task.subject())))
    components.append('END:VTODO')  # pylint: disable=W0511
    if doFold:
        return fold(components)
    return '\r\n'.join(components) + '\r\n'


def VCalFromEffort(effort, encoding=True, doFold=True):
    encoding = ';ENCODING=QUOTED-PRINTABLE;CHARSET=UTF-8' if encoding else ''
    quote = quoteString if encoding else lambda s: s
    components = []
    components.append('BEGIN:VEVENT')
    components.append('UID:%s' % effort.id().encode('UTF-8'))
    components.append('SUMMARY%s:%s'%(encoding, quote(effort.subject())))
    components.append('DESCRIPTION%s:%s'%(encoding, quote(effort.description())))
    components.append('DTSTART:%s'%fmtDateTime(effort.getStart()))
    if effort.getStop():
        components.append('DTEND:%s'%fmtDateTime(effort.getStop()))
    components.append('END:VEVENT')
    if doFold:
        return fold(components)
    return '\r\n'.join(components) + '\r\n'


def VNoteFromNote(note, encoding=True, doFold=True):
    encoding = ';ENCODING=QUOTED-PRINTABLE;CHARSET=UTF-8' if encoding else ''
    quote = quoteString if encoding else lambda s: s
    components = []
    components.append('BEGIN:VNOTE')
    components.append('X-IRMC-LUID: %s' % note.id().encode('UTF-8'))
    components.append('SUMMARY%s: %s' % (encoding, quote(note.subject())))
    components.append('BODY%s:%s' % (encoding, quote(note.description())))
    components.append('END:VNOTE')
    if note.categories(recursive=True, upwards=True):
        categories = ','.join([quote(unicode(c)) for c in note.categories(recursive=True, upwards=True)])
        components.append('CATEGORIES%s:%s'%(encoding, categories))
    if doFold:
        return fold(components)
    return '\r\n'.join(components) + '\r\n'

#}

def fold(components, linewidth=75, eol='\r\n', indent=' '):
    lines = []
    # The iCalendar standard doesn't clearly state whether the maximum line 
    # width includes the indentation or not. We keep on the safe side:
    indentedlinewidth = linewidth - len(indent)
    for component in components:
        componentLines = component.split('\n')
        firstLine = componentLines[0]
        firstLine, remainderFirstLine = firstLine[:linewidth], firstLine[linewidth:]
        lines.append(firstLine)
        while remainderFirstLine:
            nextLine, remainderFirstLine = remainderFirstLine[:indentedlinewidth], remainderFirstLine[indentedlinewidth:]
            lines.append(indent + nextLine)
        for line in componentLines[1:]:
            nextLine, remainder = line[:linewidth], line[linewidth:]
            lines.append(indent + nextLine)
            while remainder:
                nextLine, remainder = remainder[:indentedlinewidth], remainder[indentedlinewidth:]
                lines.append(indent + nextLine)
    return eol.join(lines) + eol if lines else ''



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx

from notifier import AbstractNotifier


class KNotifyNotifier(AbstractNotifier):
    def __init__(self):
        super(KNotifyNotifier, self).__init__()

        self.__factory = None
        try:
            import pcop
            import pydcop
        except ImportError:
            pass
        else:
            if pydcop.anyAppCalled('knotify') is not None:
                self.__factory = pydcop.anyAppCalled

    def getName(self):
        return u'KNotify'

    def isAvailable(self):
        return self.__factory is not None

    def notify(self, title, summary, bitmap, **kwargs):
        # KNotify does not support icons

        coding = wx.Locale_GetSystemEncodingName()

        kn = self.__factory('knotify')
        kn.default.notify('reminder',
                          title.encode(coding),
                          summary.encode(coding),
                          '', '',
                          16, 0, kwargs.get('windowId', 0))


AbstractNotifier.register(KNotifyNotifier())



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from taskcoachlib import persistence
from taskcoachlib.domain import task, note, category, effort, date


class TaskFileTestCase(test.TestCase):
    def setUp(self):
        self.taskFile = persistence.TaskFile()
        self.task = task.Task(subject='Subject')

    def tearDown(self):
        super(TaskFileTestCase, self).tearDown()
        self.taskFile.close()
        self.taskFile.stop()

    def testTaskIsDirtyAfterEditingSubject(self):
        self.taskFile.tasks().append(self.task)
        self.failUnless(task.Task.STATUS_NEW, self.task.getStatus())
        self.task.setSubject('New subject')
        self.failUnless(task.Task.STATUS_CHANGED, self.task.getStatus())

    def testNoteIsDirtyAfterEditingSubject(self):
        aNote = note.Note(subject='Subject')
        self.taskFile.notes().append(aNote)
        self.failUnless(note.Note.STATUS_NEW, aNote.getStatus())
        aNote.setSubject('New subject')
        self.failUnless(note.Note.STATUS_CHANGED, aNote.getStatus())

    def testCategoryIsDirtyAfterEditingSubject(self):
        aCategory = category.Category(subject='Subject')
        self.taskFile.categories().append(aCategory)
        self.failUnless(category.Category.STATUS_NEW, aCategory.getStatus())
        aCategory.setSubject('New subject')
        self.failUnless(category.Category.STATUS_CHANGED, aCategory.getStatus())

    def testEffortIsDirtyAfterEditingStart(self):
        self.taskFile.tasks().append(self.task)
        anEffort = effort.Effort(self.task)
        self.task.addEffort(anEffort)
        self.failUnless(effort.Effort.STATUS_NEW, anEffort.getStatus())
        anEffort.setStart(date.DateTime(2000,1,1,10,0,0))
        self.failUnless(effort.Effort.STATUS_CHANGED, anEffort.getStatus())



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from taskcoachlib import patterns


class CompositeTest(test.TestCase):
    def setUp(self):
        self.composite = patterns.Composite()
        self.child = patterns.Composite()
        
    def testNoParentByDefault(self):
        self.failIf(self.composite.parent())
        
    def testNoChildrenByDefault(self):
        self.failIf(self.composite.children())
        
    def testNoRecursiveChildrenByDefault(self):
        self.failIf(self.composite.children(recursive=True))
    
    def testNoAncestorsByDefault(self):
        self.assertEqual([], self.composite.ancestors())
        
    def testAddChild(self):
        self.composite.addChild(self.child)
        self.assertEqual([self.child], self.composite.children())
        
    def testAddChild_SetsParentOfChild(self):
        self.composite.addChild(self.child)
        self.assertEqual(self.composite, self.child.parent())
        
    def testRemoveChild(self):
        self.composite.addChild(self.child)
        self.composite.removeChild(self.child)
        self.assertEqual([], self.composite.children())
        
    def testRemoveChild_DoesNotResetParentOfChild(self):
        self.composite.addChild(self.child)
        self.composite.removeChild(self.child)
        self.assertEqual(self.composite, self.child.parent())
    
    def testCreateWithChildren(self):
        objectWithChildren = patterns.Composite(children=[self.child])
        self.assertEqual([self.child], objectWithChildren.children())
        
    def testCreateWithParent(self):
        objectWithParent = patterns.Composite(parent=self.composite)
        self.assertEqual(self.composite, objectWithParent.parent())
        
    def testCreateWithParentDoesNotAddObjectToParent(self):
        patterns.Composite(parent=self.composite)
        self.failIf(self.composite.children())
        
    def testChildrenRecursive_WithoutGrandChildren(self):
        self.composite.addChild(self.child)
        self.assertEqual([self.child], 
                         self.composite.children(recursive=True))

    def testChildrenRecursive(self):
        self.composite.addChild(self.child)
        grandChild = patterns.Composite()
        self.child.addChild(grandChild)
        self.assertEqual([self.child, grandChild], 
                         self.composite.children(recursive=True))
        
    def testAncestorsWithTwoGenerations(self):
        self.composite.addChild(self.child)
        self.assertEqual([self.composite], self.child.ancestors())
        
    def testAncestorsWithThreeGenerations(self):
        grandChild = patterns.Composite()
        self.composite.addChild(self.child)
        self.child.addChild(grandChild)
        self.assertEqual([self.composite, self.child], grandChild.ancestors())
        
    def testCompositeIsItsOnlyFamilyByDefault(self):
        self.assertEqual([self.composite], self.composite.family())
        
    def testParentIsPartOfFamily(self):
        self.composite.addChild(self.child)
        self.assertEqual([self.composite, self.child], self.child.family())
        
    def testChildIsPartOfFamily(self):
        self.composite.addChild(self.child)
        self.assertEqual([self.composite, self.child], self.composite.family())
        
    def testSiblingsWithoutParentIsEmpty(self):
        self.failIf(self.composite.siblings())
        
    def testSiblingsWithoutSiblings(self):
        self.composite.addChild(self.child)
        self.failIf(self.child.siblings())
        
    def testSiblingsWithOneSibling(self):
        self.composite.addChild(self.child)
        child2 = patterns.Composite()
        self.composite.addChild(child2)
        self.assertEqual([child2], self.child.siblings())
        
    def testNewChild_HasCorrectParent(self):
        child = self.composite.newChild()
        self.assertEqual(self.composite, child.parent())
        
    def testNewChild_NotInParentsChildren(self):
        child = self.composite.newChild()
        self.failIf(child in self.composite.children())
        
    def testGetState(self):
        self.assertEqual(dict(children=[], parent=None), 
                         self.composite.__getstate__())
        
    def testGetState_WithChildren(self):
        self.composite.addChild(self.child)
        state = self.composite.__getstate__()
        self.assertEqual(dict(children=[state['children'][0]], parent=None), 
                         state)
        
    def testGetState_WithParent(self):
        self.composite.addChild(self.child)
        self.assertEqual(dict(children=[], parent=self.composite),
                         self.child.__getstate__())
                
    def testSetState_Parent(self):
        state = self.composite.__getstate__()
        self.composite.setParent(self.child)
        self.composite.__setstate__(state)
        self.assertEqual(None, self.composite.parent())
        
    def testSetState_Children(self):
        state = self.composite.__getstate__()
        self.composite.addChild(self.child)
        self.composite.__setstate__(state)
        self.assertEqual([], self.composite.children())
        
    def testCopy(self):
        copy = self.composite.copy()
        self.assertEqual(copy.children(), self.composite.children())
        
    def testCopy_AddChildrenAfterCopy(self):
        copy = self.composite.copy()
        self.composite.addChild(self.child)
        self.failIf(self.child in copy.children())
        
    def testCopy_WithChildren(self):
        self.composite.addChild(self.child)
        copy = self.composite.copy()
        self.assertEqual(1, len(copy.children()))
    
    def testCopy_WithChildren_ParentOfCopiedChildrenIsNewComposite(self):
        self.composite.addChild(self.child)
        copy = self.composite.copy()
        self.assertEqual(copy, copy.children()[0].parent())
    
    def testCopy_WithParent(self):
        self.composite.addChild(self.child)
        copy = self.child.copy()
        self.assertEqual(self.child.parent(), copy.parent())
        
    def testCopy_WithChildren_DoesNotCreateExtraChildrenForOriginal(self):
        self.composite.addChild(self.child)
        self.composite.copy()
        self.assertEqual(1, len(self.composite.children()))

        
class ObservableCompositeTest(test.TestCase):
    def setUp(self):
        self.composite = patterns.ObservableComposite()
        self.child = patterns.ObservableComposite()
        
    def testAddChild(self):
        eventType = self.composite.addChildEventType()
        self.registerObserver(eventType)    
        self.composite.addChild(self.child)
        self.assertEqual([patterns.Event(eventType, self.composite, 
            self.child)], self.events)
            
    def testRemoveChild(self):
        eventType = self.composite.removeChildEventType()
        self.registerObserver(eventType)    
        self.composite.addChild(self.child)
        self.composite.removeChild(self.child)
        self.assertEqual([patterns.Event(eventType, self.composite,
            self.child)], self.events)

    def testModificationEventTypes(self):
        self.assertEqual([self.composite.addChildEventType(), 
                          self.composite.removeChildEventType()], 
                          self.composite.modificationEventTypes())
        
    def testSetState(self):
        self.composite.addChild(self.child)
        state = self.composite.__getstate__()
        self.registerObserver(self.composite.removeChildEventType())
        self.registerObserver(self.composite.addChildEventType())
        self.composite.__setstate__(state)
        self.assertEqual([], self.events)

            
class CompositeCollectionTest(test.TestCase):
    def setUp(self):
        self.composite = patterns.ObservableComposite()
        self.composite2 = patterns.ObservableComposite()
        self.collection = patterns.CompositeList()

    def testInitialSize(self):
        self.assertEqual(0, len(self.collection))
        
    def testCreateWithInitialContent(self):
        collection = patterns.CompositeList([self.composite])
        self.assertEqual([self.composite], collection)

    def testRootItems_NoItems(self):
        self.assertEqual([], self.collection.rootItems())
        
    def testRootItems_OneRootItem(self):
        self.collection.append(self.composite)
        self.assertEqual([self.composite], self.collection.rootItems())
        
    def testRootItems_MultipleRootItems(self):
        self.collection.extend([self.composite, self.composite2])
        self.assertEqualLists([self.composite, self.composite2], 
                         self.collection.rootItems())
        
    def testRootItems_RootAndChildItems(self):
        self.composite.addChild(self.composite2)
        self.collection.extend([self.composite, self.composite2])
        self.assertEqual([self.composite], self.collection.rootItems())
        
    def testAddChild(self):
        self.collection.extend([self.composite, self.composite2])
        self.composite.addChild(self.composite2)
        self.assertEqual([self.composite], self.collection.rootItems())
    
    def testAddRootWithChildItems(self):
        self.composite.addChild(self.composite2)
        self.collection.append(self.composite)
        self.assertEqualLists([self.composite, self.composite2], 
                              self.collection)
        
    def testAddRootWithChildItems_AddAllAtOnce(self):
        self.composite.addChild(self.composite2)
        self.collection.extend([self.composite, self.composite2])
        self.assertEqualLists([self.composite, self.composite2], 
                              self.collection)

    def testAddRootWithChildItems_DoesNotAddChildToParent(self):
        self.composite.addChild(self.composite2)
        self.collection.extend([self.composite, self.composite2])
        self.assertEqual([self.composite2], self.composite.children())
        
    def testAddCompositeWithParentAddsItToParent(self):
        self.collection.append(self.composite)
        self.composite2.setParent(self.composite)
        self.collection.append(self.composite2)
        self.assertEqual([self.composite2], self.composite.children())

    def testAddCompositeWithParentTriggersNotificationByParent(self):
        self.registerObserver(self.composite.addChildEventType())
        self.collection.append(self.composite)
        self.composite2.setParent(self.composite)
        self.collection.append(self.composite2)
        expectedEvent = patterns.Event(self.composite.addChildEventType(),
                                       self.composite, self.composite2)
        self.assertEqual([expectedEvent], self.events)
    
    def testRemoveChildFromCollectionRemovesChildFromParent(self):
        self.collection.extend([self.composite, self.composite2])
        self.composite.addChild(self.composite2)
        self.collection.remove(self.composite2)
        self.failIf(self.composite.children())
        
    def testRemoveChildFromCollectionTriggersNotificationByParent(self):
        self.registerObserver(self.composite.removeChildEventType())
        self.collection.extend([self.composite, self.composite2])
        self.composite.addChild(self.composite2)
        self.collection.remove(self.composite2)
        expectedEvent = patterns.Event(self.composite.removeChildEventType(),
                                       self.composite, self.composite2)
        self.assertEqual([expectedEvent], self.events)

    def testRemoveCompositeWithChildRemovesChildToo(self):
        self.composite.addChild(self.composite2)
        grandChild = patterns.ObservableComposite()
        self.composite2.addChild(grandChild)
        self.collection.append(self.composite)
        self.collection.remove(self.composite2)
        self.assertEqual([self.composite], self.collection)

    def testRemoveCompositeAndChildRemovesBoth(self):
        self.composite.addChild(self.composite2)
        grandChild = patterns.ObservableComposite()
        self.composite2.addChild(grandChild)
        self.collection.append(self.composite)
        self.collection.removeItems([self.composite2, grandChild])
        self.assertEqual([self.composite], self.collection)
        
    def testRemoveChildWithChildren_CollectionNotificationContainsParentAndChild(self):
        self.registerObserver(self.collection.removeItemEventType())
        self.composite.addChild(self.composite2)
        grandChild = patterns.ObservableComposite()
        self.composite2.addChild(grandChild)
        self.collection.append(self.composite)
        self.collection.remove(self.composite2)
        self.assertEqualLists([self.composite2, grandChild],
                              self.events[0].values(type=self.collection.removeItemEventType()))
    
    def testRemoveCompositeWithChildrenDoesNotBreakParentChildRelation(self):
        self.composite.addChild(self.composite2)
        self.collection.append(self.composite)
        self.collection.remove(self.composite)
        self.assertEqual([self.composite2], self.composite.children())
        
    def testRemoveChildAndThenAddingItAddsItToPreviousParentToo(self):
        self.composite.addChild(self.composite2)
        self.collection.append(self.composite)
        self.collection.remove(self.composite2)
        self.collection.append(self.composite2)
        self.assertEqual([self.composite2], self.composite.children())
        
    def testRemoveCompositeNotInCollection(self):
        self.collection.remove(self.composite)
        
        



# -*- coding: utf-8 -*-

'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>
Copyright (C) 2008 Rob McMullen <rob.mcmullen@gmail.com>
Copyright (C) 2008 Thomas Sonne Olesen <tpo@sonnet.dk>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx
from taskcoachlib import patterns, widgets, command, render
from taskcoachlib.i18n import _
from taskcoachlib.gui import uicommand, toolbar, artprovider
from taskcoachlib.thirdparty import hypertreelist
from taskcoachlib.thirdparty.pubsub import pub
from taskcoachlib.widgets import ToolTipMixin
import mixin


class Viewer(patterns.Observer, wx.Panel):
    ''' A Viewer shows domain objects (e.g. tasks or efforts) by means of a 
        widget (e.g. a ListCtrl or a TreeListCtrl).'''
    
    __metaclass__ = patterns.NumberedInstances
    defaultTitle = 'Subclass responsibility'
    defaultBitmap = 'Subclass responsibility'
    viewerImages = artprovider.itemImages
    
    def __init__(self, parent, taskFile, settings, *args, **kwargs):
        super(Viewer, self).__init__(parent, -1)
        self.parent = parent
        self.taskFile = taskFile
        self.settings = settings
        self.__settingsSection = kwargs.pop('settingsSection')
        self.__freezeCount = 0
        # The how maniest of this viewer type are we? Used for settings
        self.__instanceNumber = kwargs.pop('instanceNumber')
        self.__use_separate_settings_section = kwargs.pop('use_separate_settings_section', 
                                                          True)
        # Selection cache:
        self.__curselection = [] 
        # Flag so that we don't notify observers while we're selecting all items
        self.__selectingAllItems = False
        # Popup menus we have to destroy before closing the viewer to prevent 
        # memory leakage:
        self._popupMenus = []
        # What are we presenting:
        self.__presentation = self.createSorter(self.createFilter(self.domainObjectsToView()))
        # The widget used to present the presentation:
        self.widget = self.createWidget()
        self.widget.SetBackgroundColour(wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOW))
        self.toolbar = toolbar.ToolBar(self, settings, (16, 16))
        self.initLayout()
        self.registerPresentationObservers()
        self.refresh()

        pub.subscribe(self.onBeginIO, 'taskfile.aboutToRead')
        pub.subscribe(self.onBeginIO, 'taskfile.aboutToClear')
        pub.subscribe(self.onBeginIO, 'taskfile.aboutToSave')
        pub.subscribe(self.onEndIO, 'taskfile.justRead')
        pub.subscribe(self.onEndIO, 'taskfile.justCleared')
        pub.subscribe(self.onEndIO, 'taskfile.justSaved')

        if isinstance(self.widget, ToolTipMixin):
            pub.subscribe(self.onShowTooltipsChanged, 'settings.view.descriptionpopups')
            self.widget.SetToolTipsEnabled(settings.getboolean('view', 'descriptionpopups'))

        wx.CallAfter(self.__DisplayBalloon)

    def __DisplayBalloon(self):
        # AuiFloatingFrame is instantiated from framemanager, we can't derive it from BalloonTipManager
        if self.toolbar.IsShownOnScreen() and hasattr(wx.GetTopLevelParent(self), 'AddBalloonTip'):
            wx.GetTopLevelParent(self).AddBalloonTip(self.settings, 'customizabletoolbars', self.toolbar,
                title=_('Toolbars are customizable'),
                getRect=lambda: self.toolbar.GetToolRect(self.toolbar.getToolIdByCommand('EditToolBarPerspective')),
                message=_('''Click on the gear icon on the right to add buttons and rearrange them.'''))

    def onShowTooltipsChanged(self, value):
        self.widget.SetToolTipsEnabled(value)

    def onBeginIO(self, taskFile):
        self.__freezeCount += 1
        self.__presentation.freeze()

    def onEndIO(self, taskFile):
        self.__freezeCount -= 1
        self.__presentation.thaw()
        if self.__freezeCount == 0:
            self.refresh()

    def activate(self):
        pass

    def domainObjectsToView(self):
        ''' Return the domain objects that this viewer should display. For
            global viewers this will be part of the task file, 
            e.g. self.taskFile.tasks(), for local viewers this will be a list
            of objects passed to the viewer constructor. '''
        raise NotImplementedError
    
    def registerPresentationObservers(self):
        self.removeObserver(self.onPresentationChanged)
        self.registerObserver(self.onPresentationChanged, 
                              eventType=self.presentation().addItemEventType(),
                              eventSource=self.presentation())
        self.registerObserver(self.onPresentationChanged, 
                              eventType=self.presentation().removeItemEventType(),
                              eventSource=self.presentation())
        self.registerObserver(self.onNewItem, eventType='newitem')
               
    def detach(self):
        ''' Should be called by viewer.container before closing the viewer '''
        observers = [self, self.presentation()]
        observable = self.presentation()
        while True:
            try:
                observable = observable.observable()
            except AttributeError:
                break
            else:
                observers.append(observable)
        for observer in observers:
            try:
                observer.removeInstance()
            except AttributeError:
                pass  # Ignore observables that are not an observer themselves

        for popupMenu in self._popupMenus:
            try:
                popupMenu.clearMenu()
                popupMenu.Destroy()
            except wx.PyDeadObjectError:
                pass

        pub.unsubscribe(self.onBeginIO, 'taskfile.aboutToRead')
        pub.unsubscribe(self.onBeginIO, 'taskfile.aboutToClear')
        pub.unsubscribe(self.onBeginIO, 'taskfile.aboutToSave')
        pub.unsubscribe(self.onEndIO, 'taskfile.justRead')
        pub.unsubscribe(self.onEndIO, 'taskfile.justCleared')
        pub.unsubscribe(self.onEndIO, 'taskfile.justSaved')

        self.presentation().detach()
        self.toolbar.detach()

    def viewerStatusEventType(self):
        return 'viewer%s.status' % id(self)
    
    def sendViewerStatusEvent(self):
        pub.sendMessage(self.viewerStatusEventType(), viewer=self)
    
    def statusMessages(self):
        return '', ''
    
    def title(self):
        return self.settings.get(self.settingsSection(), 'title') or self.defaultTitle
    
    def setTitle(self, title):
        titleToSaveInSettings = '' if title == self.defaultTitle else title
        self.settings.set(self.settingsSection(), 'title', titleToSaveInSettings)
        self.parent.setPaneTitle(self, title)
        self.parent.manager.Update()

    def initLayout(self):
        self._sizer = wx.BoxSizer(wx.VERTICAL)  # pylint: disable=W0201
        self._sizer.Add(self.toolbar, flag=wx.EXPAND)
        self._sizer.Add(self.widget, proportion=1, flag=wx.EXPAND)
        self.SetSizerAndFit(self._sizer)
    
    def createWidget(self, *args):
        raise NotImplementedError
    
    def createImageList(self):
        size = (16, 16)
        imageList = wx.ImageList(*size)  # pylint: disable=W0142
        self.imageIndex = {}  # pylint: disable=W0201
        for index, image in enumerate(self.viewerImages):
            try:
                imageList.Add(wx.ArtProvider_GetBitmap(image, wx.ART_MENU, size))
            except:
                print image
                raise
            self.imageIndex[image] = index
        return imageList

    def getWidget(self):
        return self.widget
    
    def SetFocus(self, *args, **kwargs):
        try:
            self.widget.SetFocus(*args, **kwargs)
        except wx.PyDeadObjectError:
            pass
            
    def createSorter(self, collection):
        ''' This method can be overridden to decorate the presentation with a 
            sorter. '''
        return collection
        
    def createFilter(self, collection):
        ''' This method can be overridden to decorate the presentation with a 
            filter. '''
        return collection

    def onAttributeChanged(self, newValue, sender):  # pylint: disable=W0613
        if self:
            self.refreshItems(sender)
        
    def onAttributeChanged_Deprecated(self, event):
        self.refreshItems(*event.sources())
        
    def onNewItem(self, event):
        self.select([item for item in event.values() if item in self.presentation()]) 
        
    def onPresentationChanged(self, event):  # pylint: disable=W0613
        ''' Whenever our presentation is changed (items added, items removed) 
            the viewer refreshes itself. '''
        def itemsRemoved():
            return event.type() == self.presentation().removeItemEventType()
        
        def allItemsAreSelected():
            return set(self.__curselection).issubset(set(event.values()))
        
        self.refresh()
        if itemsRemoved() and allItemsAreSelected(): 
            self.selectNextItemsAfterRemoval(event.values())
        self.updateSelection(sendViewerStatusEvent=False)
        self.sendViewerStatusEvent()
        
    def selectNextItemsAfterRemoval(self, removedItems):
        raise NotImplementedError        
        
    def onSelect(self, event=None):  # pylint: disable=W0613
        ''' The selection of items in the widget has been changed. Notify 
            our observers. '''
        if self.IsBeingDeleted() or self.__selectingAllItems:
            # Some widgets change the selection and send selection events when 
            # deleting all items as part of the Destroy process. Ignore.
            return
        # Be sure all wx events are handled before we update our selection 
        # cache and notify our observers:
        wx.CallAfter(self.updateSelection)

    def updateSelection(self, sendViewerStatusEvent=True):
        newSelection = self.widget.curselection()
        if newSelection != self.__curselection:
            self.__curselection = newSelection
            if sendViewerStatusEvent:
                self.sendViewerStatusEvent()

    def freeze(self):
        self.widget.Freeze()

    def thaw(self):
        self.widget.Thaw()

    def refresh(self):
        if self and not self.__freezeCount:
            self.widget.RefreshAllItems(len(self.presentation()))
    
    def refreshItems(self, *items):
        if not self.__freezeCount:
            items = [item for item in items if item in self.presentation()]
            self.widget.RefreshItems(*items)  # pylint: disable=W0142
        
    def select(self, items):
        self.__curselection = items
        self.widget.select(items)
        
    def curselection(self):
        ''' Return a list of items (domain objects) currently selected in our
            widget. '''
        return self.__curselection
        
    def curselectionIsInstanceOf(self, class_):
        ''' Return whether all items in the current selection are instances of
            class_. Can be overridden in subclasses that show only one type
            of items to simply check the class. '''
        return all(isinstance(item, class_) for item in self.curselection())

    def isselected(self, item):
        ''' Returns True if the given item is selected. See
            L{EffortViewer} for an explanation of why this may be
            different than 'if item in viewer.curselection()'. '''
        return item in self.curselection()

    def select_all(self):
        ''' Select all items in the presentation. Since some of the widgets we
            use may send events for each individual item (!) we stop processing
            selection events while we select all items. '''
        self.__selectingAllItems = True
        self.widget.select_all()
        # Use CallAfter to make sure we start processing selection events 
        # after all selection events have been fired (and ignored):
        wx.CallAfter(self.endOfSelectAll)
        
    def endOfSelectAll(self):
        self.__curselection = self.presentation()
        self.__selectingAllItems = False
        # Pretend we received one selection event for the select_all() call:
        self.onSelect()

    def clear_selection(self):
        self.__curselection = []
        self.widget.clear_selection()
        
    def size(self):
        return self.widget.GetItemCount()
    
    def presentation(self):
        ''' Return the domain objects that this viewer is currently 
            displaying. '''
        return self.__presentation
        
    def setPresentation(self, presentation):
        ''' Change the presentation of the viewer. '''
        self.__presentation = presentation
    
    def widgetCreationKeywordArguments(self):
        return {}

    def isViewerContainer(self):
        return False
    
    def isShowingTasks(self): 
        return False

    def isShowingEffort(self): 
        return False
    
    def isShowingCategories(self):
        return False
    
    def isShowingNotes(self):
        return False

    def isShowingAttachments(self):
        return False

    def visibleColumns(self):
        return [widgets.Column('subject', _('Subject'))]
    
    def bitmap(self):
        ''' Return the bitmap that represents this viewer. Used for the 
            'Viewer->New viewer' menu item, for example. '''
        return self.defaultBitmap  # Class attribute of concrete viewers
    
    def settingsSection(self):
        ''' Return the settings section of this viewer. '''
        section = self.__settingsSection
        if self.__use_separate_settings_section and self.__instanceNumber > 0:
            # We're not the first viewer of our class, so we need a different
            # settings section than the default one.
            section += str(self.__instanceNumber)
            if not self.settings.has_section(section):
                # Our section does not exist yet. Create it and copy the 
                # settings from the previous section as starting point. We're 
                # copying from the previous section instead of the default
                # section so that when the user closes a viewer and then opens
                # a new one, the settings of that closed viewer are reused. 
                self.settings.add_section(section, 
                    copyFromSection=self.previousSettingsSection())
        return section
        
    def previousSettingsSection(self):
        ''' Return the settings section of the previous viewer of this 
            class. '''
        previousSectionNumber = self.__instanceNumber - 1
        while previousSectionNumber > 0:
            previousSection = self.__settingsSection + str(previousSectionNumber)
            if self.settings.has_section(previousSection):
                return previousSection
            previousSectionNumber -= 1
        return self.__settingsSection
    
    def hasModes(self):
        return False
    
    def getModeUICommands(self):
        return []
    
    def isSortable(self):
        return False

    def getSortUICommands(self):
        return []
    
    def isSearchable(self):
        return False
        
    def hasHideableColumns(self):
        return False
    
    def getColumnUICommands(self):
        return []

    def isFilterable(self):
        return False
    
    def getFilterUICommands(self):
        return []
    
    def supportsRounding(self):
        return False
    
    def getRoundingUICommands(self):
        return []
    
    def createToolBarUICommands(self):
        ''' UI commands to put on the toolbar of this viewer. '''
        table = wx.AcceleratorTable([(wx.ACCEL_CMD, ord('X'), wx.ID_CUT),
                                     (wx.ACCEL_CMD, ord('C'), wx.ID_COPY),
                                     (wx.ACCEL_CMD, ord('V'), wx.ID_PASTE),
                                     (wx.ACCEL_NORMAL, wx.WXK_RETURN, wx.ID_EDIT),
                                     (wx.ACCEL_CTRL, wx.WXK_DELETE, wx.ID_DELETE)])
        self.SetAcceleratorTable(table)
        
        clipboardToolBarUICommands = self.createClipboardToolBarUICommands()
        creationToolBarUICommands = self.createCreationToolBarUICommands()
        editToolBarUICommands = self.createEditToolBarUICommands() 
        actionToolBarUICommands = self.createActionToolBarUICommands()
        modeToolBarUICommands = self.createModeToolBarUICommands()
        
        def separator(uiCommands, *otherUICommands):
            return (None,) if (uiCommands and any(otherUICommands)) else ()
        
        clipboardSeparator = separator(clipboardToolBarUICommands, creationToolBarUICommands,
                                       editToolBarUICommands, actionToolBarUICommands,
                                       modeToolBarUICommands)
        creationSeparator = separator(creationToolBarUICommands, editToolBarUICommands,
                                      actionToolBarUICommands, modeToolBarUICommands)
        editSeparator = separator(editToolBarUICommands, actionToolBarUICommands,
                                  modeToolBarUICommands)
        actionSeparator = separator(actionToolBarUICommands, modeToolBarUICommands)
        
        return clipboardToolBarUICommands + clipboardSeparator + \
            creationToolBarUICommands + creationSeparator + \
            editToolBarUICommands + editSeparator + \
            actionToolBarUICommands + actionSeparator + \
            modeToolBarUICommands

    def getToolBarPerspective(self):
        return self.settings.get(self.settingsSection(), 'toolbarperspective')

    def saveToolBarPerspective(self, perspective):
        self.settings.set(self.settingsSection(), 'toolbarperspective', perspective)

    def createClipboardToolBarUICommands(self):
        ''' UI commands for manipulating the clipboard (cut, copy, paste). '''
        cutCommand = uicommand.EditCut(viewer=self)
        copyCommand = uicommand.EditCopy(viewer=self)
        pasteCommand = uicommand.EditPaste()
        cutCommand.bind(self, wx.ID_CUT)
        copyCommand.bind(self, wx.ID_COPY)
        pasteCommand.bind(self, wx.ID_PASTE)
        return cutCommand, copyCommand, pasteCommand
    
    def createCreationToolBarUICommands(self):
        ''' UI commands for creating new items. '''
        return ()
    
    def createEditToolBarUICommands(self):
        ''' UI commands for editing items. '''
        editCommand = uicommand.Edit(viewer=self)
        self.deleteUICommand = uicommand.Delete(viewer=self)  # For unittests pylint: disable=W0201
        editCommand.bind(self, wx.ID_EDIT)
        self.deleteUICommand.bind(self, wx.ID_DELETE)
        return editCommand, self.deleteUICommand
    
    def createActionToolBarUICommands(self):
        ''' UI commands for actions. '''
        return ()
    
    def createModeToolBarUICommands(self):
        ''' UI commands for mode switches (e.g. list versus tree mode). '''
        return ()
    
    def newItemDialog(self, *args, **kwargs):
        bitmap = kwargs.pop('bitmap')
        newItemCommand = self.newItemCommand(*args, **kwargs)
        newItemCommand.do()
        return self.editItemDialog(newItemCommand.items, bitmap, 
                                   items_are_new=True)

    def newSubItemDialog(self, bitmap):
        newSubItemCommand = self.newSubItemCommand()
        newSubItemCommand.do()
        return self.editItemDialog(newSubItemCommand.items, bitmap, 
                                   items_are_new=True)   
    
    def editItemDialog(self, items, bitmap, columnName='', items_are_new=False):
        Editor = self.itemEditorClass()
        return Editor(wx.GetTopLevelParent(self), items, 
                      self.settings, self.presentation(), self.taskFile, 
                      bitmap=bitmap, columnName=columnName, 
                      items_are_new=items_are_new)
        
    def itemEditorClass(self):
        raise NotImplementedError
    
    def newItemCommand(self, *args, **kwargs):
        return self.newItemCommandClass()(self.presentation(), *args, **kwargs)

    def newItemCommandClass(self):
        raise NotImplementedError
    
    def newSubItemCommand(self):
        return self.newSubItemCommandClass()(self.presentation(), 
                                             self.curselection())

    def newSubItemCommandClass(self):
        raise NotImplementedError

    def deleteItemCommand(self):
        return self.deleteItemCommandClass()(self.presentation(), self.curselection())

    def deleteItemCommandClass(self):
        return command.DeleteCommand

    def cutItemCommand(self):
        return self.cutItemCommandClass()(self.presentation(), self.curselection())

    def cutItemCommandClass(self):
        return command.CutCommand

    def onEditSubject(self, item, newValue):
        command.EditSubjectCommand(items=[item], newValue=newValue).do()
        
    def onEditDescription(self, item, newValue):
        command.EditDescriptionCommand(items=[item], newValue=newValue).do()

    def getItemTooltipData(self, item):
        lines = [line.rstrip('\r') for line in item.description().split('\n')] 
        return [(None, lines)] if lines and lines != [''] else [] 


class CategorizableViewerMixin(object):
    def getItemTooltipData(self, item):
        return [('folder_blue_arrow_icon', [u', '.join(sorted([cat.subject() for cat in item.categories()]))] if item.categories() else [])] + \
            super(CategorizableViewerMixin, self).getItemTooltipData(item)


class WithAttachmentsViewerMixin(object):
    def getItemTooltipData(self, item):
        return [('paperclip_icon', sorted([unicode(attachment) for attachment in item.attachments()]))] + \
          super(WithAttachmentsViewerMixin, self).getItemTooltipData(item)


class ListViewer(Viewer):  # pylint: disable=W0223
    def isTreeViewer(self):
        return False

    def visibleItems(self):
        ''' Iterate over the items in the presentation. '''
        for item in self.presentation():
            yield item
    
    def getItemWithIndex(self, index):
        return self.presentation()[index]
            
    def getIndexOfItem(self, item):
        return self.presentation().index(item)

    def selectNextItemsAfterRemoval(self, removedItems):
        pass  # Done automatically by list controls        
        

class TreeViewer(Viewer):  # pylint: disable=W0223
    def __init__(self, *args, **kwargs):
        self.__selectionIndex = 0
        super(TreeViewer, self).__init__(*args, **kwargs)
        self.widget.Bind(wx.EVT_TREE_ITEM_EXPANDED, self.onItemExpanded)
        self.widget.Bind(wx.EVT_TREE_ITEM_COLLAPSED, self.onItemCollapsed)

    def onItemExpanded(self, event):
        self.__handleExpandedOrCollapsedItem(event, expanded=True)
        
    def onItemCollapsed(self, event):
        self.__handleExpandedOrCollapsedItem(event, expanded=False)
        
    def __handleExpandedOrCollapsedItem(self, event, expanded):
        event.Skip()
        treeItem = event.GetItem()
        # If we get an expanded or collapsed event for the root item, ignore it
        if treeItem == self.widget.GetRootItem():
            return
        item = self.widget.GetItemPyData(treeItem)
        item.expand(expanded, context=self.settingsSection())
            
    def expandAll(self):
        ''' Expand all items, recursively. '''
        # Since the widget does not send EVT_TREE_ITEM_EXPANDED when expanding
        # all items, we have to do the bookkeeping ourselves:
        for item in self.visibleItems():
            item.expand(True, context=self.settingsSection(), notify=False)
        self.refresh()

    def collapseAll(self):
        ''' Collapse all items, recursively. '''
        # Since the widget does not send EVT_TREE_ITEM_COLLAPSED when collapsing
        # all items, we have to do the bookkeeping ourselves:
        for item in self.visibleItems():
            item.expand(False, context=self.settingsSection(), notify=False)
        self.refresh()
                
    def isAnyItemExpandable(self):
        return self.widget.isAnyItemExpandable()

    def isAnyItemCollapsable(self):
        return self.widget.isAnyItemCollapsable()
        
    def isTreeViewer(self):
        return True
    
    def select(self, items):
        for item in items:
            self.__expandItemRecursively(item)
        self.refresh()
        super(TreeViewer, self).select(items)
        
    def __expandItemRecursively(self, item):
        parent = self.getItemParent(item)
        if parent:
            parent.expand(True, context=self.settingsSection(), notify=False)
            self.__expandItemRecursively(parent)

    def selectNextItemsAfterRemoval(self, removedItems):
        parents = [self.getItemParent(item) for item in removedItems]
        parents = [parent for parent in parents if parent in self.presentation()]
        parent = parents[0] if parents else None
        siblings = self.children(parent)
        newSelection = siblings[min(len(siblings) - 1, self.__selectionIndex)] if siblings else parent
        if newSelection:
            self.select([newSelection])
    
    def updateSelection(self, *args, **kwargs):        
        super(TreeViewer, self).updateSelection(*args, **kwargs)
        curselection = self.curselection()
        if curselection:
            siblings = self.children(self.getItemParent(curselection[0]))
            self.__selectionIndex = siblings.index(curselection[0]) if curselection[0] in siblings else 0
        else:
            self.__selectionIndex = 0
            
    def visibleItems(self):
        ''' Iterate over the items in the presentation. '''            
                            
        def yieldItemsAndChildren(items):
            sortedItems = [item for item in self.presentation() if item in items]
            for item in sortedItems:
                yield item
                children = self.children(item)
                if children:
                    for child in yieldItemsAndChildren(children):
                        yield child

        for item in yieldItemsAndChildren(self.getRootItems()):
            yield item
        
    def getRootItems(self):
        ''' Allow for overriding what the rootItems are. '''
        return self.presentation().rootItems()
            
    def getItemParent(self, item):
        ''' Allow for overriding what the parent of an item is. '''
        return item.parent()

    def getItemExpanded(self, item):
        return item.isExpanded(context=self.settingsSection())
    
    def children(self, parent=None):
        if parent:
            children = parent.children()
            if children:
                return [child for child in self.presentation() if child in children]
            else:
                return []
        else:
            return self.getRootItems()
        
    def getItemText(self, item):
        return item.subject()

            
class ViewerWithColumns(Viewer):  # pylint: disable=W0223
    def __init__(self, *args, **kwargs):
        self.__initDone = False
        self._columns = []
        self.__visibleColumns = []
        self.__columnUICommands = []
        super(ViewerWithColumns, self).__init__(*args, **kwargs)
        self.initColumns()
        self.__initDone = True
        self.refresh()
        
    def hasHideableColumns(self):
        return True

    def hasOrderingColumn(self):
        for column in self.__visibleColumns:
            if column.name() == 'ordering':
                return True
        return False

    def getColumnUICommands(self):
        if not self.__columnUICommands:
            self.__columnUICommands = self.createColumnUICommands()
        return self.__columnUICommands

    def createColumnUICommands(self):
        raise NotImplementedError
    
    def refresh(self, *args, **kwargs):
        if self and self.__initDone:
            super(ViewerWithColumns, self).refresh(*args, **kwargs)
                    
    def initColumns(self):
        for column in self.columns():
            self.initColumn(column)
        if self.hasOrderingColumn():
            self.widget.SetResizeColumn(1)
            self.widget.SetMainColumn(1)

    def initColumn(self, column):
        if column.name() in self.settings.getlist(self.settingsSection(), 
                                                  'columnsalwaysvisible'):
            show = True
        else:
            show = column.name() in self.settings.getlist(self.settingsSection(), 'columns')
            self.widget.showColumn(column, show=show)
        if show:
            self.__visibleColumns.append(column)
            self.__startObserving(column.eventTypes())
    
    def showColumnByName(self, columnName, show=True):
        for column in self.hideableColumns():
            if columnName == column.name():
                isVisibleColumn = self.isVisibleColumn(column)
                if (show and not isVisibleColumn) or \
                   (not show and isVisibleColumn):
                    self.showColumn(column, show)
                break

    def showColumn(self, column, show=True, refresh=True):
        if column.name() == 'ordering':
            self.widget.SetResizeColumn(1 if show else 0)
            self.widget.SetMainColumn(1 if show else 0)

        if show:
            self.__visibleColumns.append(column)
            # Make sure we keep the columns in the right order:
            self.__visibleColumns = [c for c in self.columns() if \
                                     c in self.__visibleColumns]
            self.__startObserving(column.eventTypes())
        else:
            self.__visibleColumns.remove(column)
            self.__stopObserving(column.eventTypes())
        self.widget.showColumn(column, show)
        self.settings.set(self.settingsSection(), 'columns', 
            str([column.name() for column in self.__visibleColumns]))
        if refresh:
            self.widget.RefreshAllItems(len(self.presentation()))

    def hideColumn(self, visibleColumnIndex):
        column = self.visibleColumns()[visibleColumnIndex]
        self.showColumn(column, show=False)
                
    def columns(self):
        return self._columns

    def selectableColumns(self):
        return self._columns

    def isVisibleColumnByName(self, columnName):
        return columnName in [column.name() for column in self.__visibleColumns]
        
    def isVisibleColumn(self, column):
        return column in self.__visibleColumns
    
    def visibleColumns(self):
        return self.__visibleColumns
        
    def hideableColumns(self):
        return [column for column in self._columns if column.name() not in \
                self.settings.getlist(self.settingsSection(), 
                                      'columnsalwaysvisible')]
                
    def isHideableColumn(self, visibleColumnIndex):
        column = self.visibleColumns()[visibleColumnIndex]
        unhideableColumns = self.settings.getlist(self.settingsSection(), 
                                                  'columnsalwaysvisible')
        return column.name() not in unhideableColumns

    def getColumnWidth(self, columnName):
        columnWidths = self.settings.getdict(self.settingsSection(),
                                             'columnwidths')
        defaultWidth = 28 if columnName == 'ordering' else hypertreelist._DEFAULT_COL_WIDTH  # pylint: disable=W0212
        return columnWidths.get(columnName, defaultWidth)

    def onResizeColumn(self, column, width):
        columnWidths = self.settings.getdict(self.settingsSection(), 'columnwidths')
        columnWidths[column.name()] = width
        self.settings.setdict(self.settingsSection(), 'columnwidths', columnWidths)

    def validateDrag(self, dropItem, dragItems, columnIndex):
        if columnIndex == -1 or self.visibleColumns()[columnIndex].name() != 'ordering':
            return None # Normal behavior

        # Ordering

        if not self.isTreeViewer():
            return True

        # Tree mode. Only allow drag if all selected items are siblings.
        if len(set([item.parent() for item in dragItems])) >= 2:
            wx.GetTopLevelParent(self).AddBalloonTip(self.settings, 'treemanualordering', self,
                title=_('Reordering in tree mode'),
                getRect=lambda: wx.Rect(0, 0, 28, 16),
                message=_('''When in tree mode, manual ordering is only possible when all selected items are siblings.'''))
            return False

        # If they are, only allow drag at the same level
        if dragItems[0].parent() != (None if dropItem is None else dropItem.parent()):
            wx.GetTopLevelParent(self).AddBalloonTip(self.settings, 'treechildrenmanualordering', self,
                title=_('Reordering in tree mode'),
                getRect=lambda: wx.Rect(0, 0, 28, 16),
                message=_('''When in tree mode, you can only put objects at the same level (parent).'''))
            return False

        return True

    def getItemText(self, item, column=None):
        if column is None:
            column = 1 if self.hasOrderingColumn() else 0
        column = self.visibleColumns()[column]
        return column.render(item)
    
    def getItemImages(self, item, column=0):
        column = self.visibleColumns()[column]
        return column.imageIndices(item)
    
    def hasColumnImages(self, column):
        return self.visibleColumns()[column].hasImages()
    
    def subjectImageIndices(self, item):
        normalIcon = item.icon(recursive=True)
        selectedIcon = item.selectedIcon(recursive=True) or normalIcon
        normalImageIndex = self.imageIndex[normalIcon] if normalIcon else -1
        selectedImageIndex = self.imageIndex[selectedIcon] if selectedIcon else -1
        return {wx.TreeItemIcon_Normal: normalImageIndex,
                wx.TreeItemIcon_Expanded: selectedImageIndex} 
            
    def __startObserving(self, eventTypes):
        for eventType in eventTypes:
            if eventType.startswith('pubsub'):
                pub.subscribe(self.onAttributeChanged, eventType)
            else:
                self.registerObserver(self.onAttributeChanged_Deprecated, 
                                      eventType=eventType)                    
        
    def __stopObserving(self, eventTypes):
        # Collect the event types that the currently visible columns are
        # interested in and make sure we don't stop observing those event types.
        eventTypesOfVisibleColumns = []
        for column in self.visibleColumns():
            eventTypesOfVisibleColumns.extend(column.eventTypes())
        for eventType in eventTypes:
            if eventType not in eventTypesOfVisibleColumns:
                if eventType.startswith('pubsub'):
                    pub.unsubscribe(self.onAttributeChanged, eventType)
                else:
                    self.removeObserver(self.onAttributeChanged_Deprecated, eventType=eventType)

    def renderCategories(self, item):
        return self.renderSubjectsOfRelatedItems(item, item.categories)        
    
    def renderSubjectsOfRelatedItems(self, item, getItems):
        subjects = []
        ownItems = getItems(recursive=False)
        if ownItems:
            subjects.append(self.renderSubjects(ownItems))
        isListViewer = not self.isTreeViewer()  # pylint: disable=E1101
        if isListViewer or self.isItemCollapsed(item):
            childItems = [theItem for theItem in getItems(recursive=True, upwards=isListViewer) if theItem not in ownItems]
            if childItems:
                subjects.append('(%s)' % self.renderSubjects(childItems))
        return ' '.join(subjects)
    
    @staticmethod
    def renderSubjects(items):
        subjects = [item.subject(recursive=True) for item in items]
        return ', '.join(sorted(subjects))
    
    @staticmethod
    def renderCreationDateTime(item, humanReadable=True):
        return render.dateTime(item.creationDateTime(), 
                               humanReadable=humanReadable)
        
    @staticmethod
    def renderModificationDateTime(item, humanReadable=True):
        return render.dateTime(item.modificationDateTime(),
                               humanReadable=humanReadable)
            
    def isItemCollapsed(self, item): 
        # pylint: disable=E1101
        # pylint: disable=E1101
        return not self.getItemExpanded(item) \
            if self.isTreeViewer() and item.children() else False


class SortableViewerWithColumns(mixin.SortableViewerMixin, ViewerWithColumns):  # pylint: disable=W0223
    def initColumn(self, column):
        super(SortableViewerWithColumns, self).initColumn(column)
        if self.isSortedBy(column.name()):
            self.widget.showSortColumn(column)
            self.showSortOrder()

    def setSortOrderAscending(self, *args, **kwargs):  # pylint: disable=W0221
        super(SortableViewerWithColumns, self).setSortOrderAscending(*args, **kwargs)
        self.showSortOrder()
        
    def sortBy(self, *args, **kwargs):  # pylint: disable=W0221
        super(SortableViewerWithColumns, self).sortBy(*args, **kwargs)
        self.showSortColumn()
        self.showSortOrder()

    def showSortColumn(self):
        for column in self.columns():
            if self.isSortedBy(column.name()):
                self.widget.showSortColumn(column)
                break

    def showSortOrder(self):
        self.widget.showSortOrder(self.imageIndex[self.getSortOrderImage()])
        
    def getSortOrderImage(self):
        return 'arrow_up_icon' if self.isSortOrderAscending() else 'arrow_down_icon'



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import sys

if sys.platform == 'win32':
    from taskcoachlib.powermgt.win32 import PowerStateMixin
elif sys.platform == 'darwin':
    from taskcoachlib.powermgt.macos import PowerStateMixin
else:
    # No way to know yet
    from taskcoachlib.powermgt.base import PowerStateMixinBase as PowerStateMixin

from taskcoachlib.powermgt.idle import IdleNotifier



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib.domain import date
import test


class CommonRecurrenceTestsMixin(object):
    def testNextDateWithInfiniteDate(self):
        self.assertEqual(date.DateTime(), self.recur(date.DateTime()))

    def testCopy(self):
        copy = self.recur.copy()
        self.assertEqual(copy, self.recur)

    def testNotEqualToNone(self):
        self.assertNotEqual(None, self.recur)

    def testSetMaxRecurrenceCount(self):
        self.recur.max = 1
        self.recur(date.Now())
        self.failIf(self.recur)
        
    def testSetMaxRecurrenceCount_GetMultipleDates(self):
        self.recur.max = 1
        self.recur(date.Now(), next=False)
        self.failUnless(self.recur)

    def testSetStopDateTime(self):
        self.recur.stop_datetime = date.Yesterday()
        self.recur(date.Now())
        self.failIf(self.recur)

    def testCount(self):
        self.assertEqual(0, self.recur.count)


class DailyRecurrenceCompareTestsMixin(object):
    def testCompareWithNone(self):
        self.failUnless(self.recur < None)
        
    def testCompareWithNoRecurrence(self):
        self.failUnless(self.recur < date.Recurrence())
      
    def testCompareWithWeeklyRecurrence(self):
        self.failUnless(self.recur < date.Recurrence('weekly'))

    def testCompareWithMonthlyRecurrence(self):
        self.failUnless(self.recur < date.Recurrence('monthly'))

    def testCompareWithYearlyRecurrence(self):
        self.failUnless(self.recur < date.Recurrence('yearly'))


class WeeklyRecurrenceCompareTestsMixin(object):
    def testCompareWithNone(self):
        self.failUnless(self.recur < None)

    def testCompareWithNoRecurrence(self):
        self.failUnless(self.recur < date.Recurrence())
      
    def testCompareWithDailyRecurrence(self):
        self.failUnless(self.recur > date.Recurrence('daily'))

    def testCompareWithMonthlyRecurrence(self):
        self.failUnless(self.recur < date.Recurrence('monthly'))

    def testCompareWithYearlyRecurrence(self):
        self.failUnless(self.recur < date.Recurrence('yearly'))


class MonthlyRecurrenceCompareTestsMixin(object):
    def testCompareWithNone(self):
        self.failUnless(self.recur < None)

    def testCompareWithNoRecurrence(self):
        self.failUnless(self.recur < date.Recurrence())
      
    def testCompareWithDailyRecurrence(self):
        self.failUnless(self.recur > date.Recurrence('daily'))

    def testCompareWithWeeklyRecurrence(self):
        self.failUnless(self.recur > date.Recurrence('weekly'))

    def testCompareWithYearlyRecurrence(self):
        self.failUnless(self.recur < date.Recurrence('yearly'))


class YearlyRecurrenceCompareTestsMixin(object):
    def testCompareWithNone(self):
        self.failUnless(self.recur < None)

    def testCompareWithNoRecurrence(self):
        self.failUnless(self.recur < date.Recurrence())
      
    def testCompareWithDailyRecurrence(self):
        self.failUnless(self.recur > date.Recurrence('daily'))

    def testCompareWithWeeklyRecurrence(self):
        self.failUnless(self.recur > date.Recurrence('weekly'))

    def testCompareWithMonthlyRecurrence(self):
        self.failUnless(self.recur > date.Recurrence('monthly'))


class NoRecurrenceTest(test.TestCase, CommonRecurrenceTestsMixin):
    def setUp(self):
        self.recur = date.Recurrence()
        
    def testNextDate(self):
        now = date.Now()
        self.assertEqual(now, self.recur(now))
        
    def testBool(self):
        self.failIf(self.recur)

    def testSetMaxRecurrenceCount_GetMultipleDates(self):
        pass
    

class DailyRecurrenceTest(test.TestCase, CommonRecurrenceTestsMixin,
                          DailyRecurrenceCompareTestsMixin):
    def setUp(self):
        self.recur = date.Recurrence('daily')
        self.now = date.Now()
                
    def testNextDate(self):
        self.assertEqual(self.now + date.ONE_DAY, self.recur(self.now))
        
    def testMultipleNextDates(self):
        self.assertEqual((self.now + date.ONE_DAY, self.now),
                         self.recur(self.now, self.now - date.ONE_DAY))
        
    def testNextDateTwice(self):
        now = self.recur(self.now - date.ONE_DAY)
        self.assertEqual(self.now + date.ONE_DAY, self.recur(now))
        
    def testCompareWithBiDailyRecurrence(self):
        self.failUnless(self.recur < date.Recurrence('daily', amount=2))
        

class BiDailyRecurrenceTest(test.TestCase, CommonRecurrenceTestsMixin,
                            DailyRecurrenceCompareTestsMixin):
    def setUp(self):
        self.recur = date.Recurrence('daily', amount=2)
        self.now = date.Now()
        
    def testEveryOtherDay(self):
        self.assertEqual(self.now + date.ONE_DAY, 
                         self.recur(self.now - date.ONE_DAY))


class TriDailyRecurrenceTest(test.TestCase, CommonRecurrenceTestsMixin,
                             DailyRecurrenceCompareTestsMixin):
    def setUp(self):
        self.recur = date.Recurrence('daily', amount=3)
        
    def testEveryThirdDay(self):
        self.assertEqual(date.DateTime(2000, 1, 4), 
                         self.recur(date.DateTime(2000, 1, 1)))
            
        
class WeeklyRecurrenceTest(test.TestCase, CommonRecurrenceTestsMixin,
                           WeeklyRecurrenceCompareTestsMixin):
    def setUp(self):
        self.January1 = date.DateTime(2000, 1, 1)
        self.January8 = date.DateTime(2000, 1, 8)
        self.January15 = date.DateTime(2000, 1, 15)
        self.recur = date.Recurrence('weekly')
        
    def testNextDate(self):
        self.assertEqual(self.January8, self.recur(self.January1))
        
    def testNextDateTwice(self):
        January8 = self.recur(self.January1)
        self.assertEqual(self.January15, self.recur(January8))

    def testCompareWithBiWeeklyRecurrence(self):
        self.failUnless(self.recur < date.Recurrence('weekly', amount=2))


class BiWeeklyRecurrenceTest(test.TestCase, CommonRecurrenceTestsMixin,
                             WeeklyRecurrenceCompareTestsMixin):
    def setUp(self):
        self.recur = date.Recurrence('weekly', amount=2)
                
    def testEveryOtherWeek(self):
        self.assertEqual(date.DateTime(2000, 1, 15, 12, 0, 0), 
                         self.recur(date.DateTime(2000, 1, 1, 12, 0, 0)))


class MonthlyRecurrenceTest(test.TestCase, CommonRecurrenceTestsMixin,
                            MonthlyRecurrenceCompareTestsMixin):
    def setUp(self):
        self.recur = date.Recurrence('monthly')
        
    def testFirstDayOf31DayMonth(self):
        self.assertEqual(date.DateTime(2000, 2, 1), 
                         self.recur(date.DateTime(2000, 1, 1)))
        
    def testFirstDayOf30DayMonth(self):
        self.assertEqual(date.DateTime(2000, 5, 1), 
                         self.recur(date.DateTime(2000, 4, 1)))
        
    def testFirstDayOfDecember(self):
        self.assertEqual(date.DateTime(2001, 1, 1), 
                         self.recur(date.DateTime(2000, 12, 1)))
        
    def testLastDayOf31DayMonth(self):
        self.assertEqual(date.DateTime(2000, 4, 30), 
                         self.recur(date.DateTime(2000, 3, 31)))
        
    def testLastDayOf30DayMonth(self):
        self.assertEqual(date.DateTime(2000, 5, 30), 
                         self.recur(date.DateTime(2000, 4, 30)))
        

class BiMontlyRecurrenceTest(test.TestCase, CommonRecurrenceTestsMixin,
                             MonthlyRecurrenceCompareTestsMixin):
    def setUp(self):
        self.recur = date.Recurrence('monthly', amount=2)
        
    def testEveryOtherMonth(self):
        self.assertEqual(date.DateTime(2000, 3, 1), 
                         self.recur(date.DateTime(2000, 1, 1)))


class MonthlySameWeekDayRecurrenceTest(test.TestCase, CommonRecurrenceTestsMixin,
                                       MonthlyRecurrenceCompareTestsMixin):
    def setUp(self):
        self.recur = date.Recurrence('monthly', sameWeekday=True)
        
    def testFirstSaturdayOfTheMonth(self):
        self.assertEqual(date.DateTime(2008, 7, 5), 
                         self.recur(date.DateTime(2008, 6, 7)))
        
    def testSecondSaturdayOfTheMonth(self):
        self.assertEqual(date.DateTime(2008, 7, 12), 
                         self.recur(date.DateTime(2008, 6, 14)))

    def testThirdSaturdayOfTheMonth(self):
        self.assertEqual(date.DateTime(2008, 7, 19), 
                         self.recur(date.DateTime(2008, 6, 21)))

    def testFourthSaturdayOfTheMonth(self):
        self.assertEqual(date.DateTime(2008, 7, 26), 
                         self.recur(date.DateTime(2008, 6, 28)))

    def testFifthSaturdayOfTheMonth_ResultsInFourthSaterdayOfTheNextMonth(self):
        self.assertEqual(date.DateTime(2008, 6, 28), 
                         self.recur(date.DateTime(2008, 5, 31)))


class BiMonthlySameWeekDayRecurrenceTest(test.TestCase, CommonRecurrenceTestsMixin,
                                         MonthlyRecurrenceCompareTestsMixin):
    def setUp(self):
        self.recur = date.Recurrence('monthly', amount=2, sameWeekday=True)
        
    def testFourthSaturdayOfTheMonth(self):
        self.assertEqual(date.DateTime(2008, 8, 23), 
                         self.recur(date.DateTime(2008, 6, 28)))


class YearlyRecurrenceTest(test.TestCase, CommonRecurrenceTestsMixin,
                           YearlyRecurrenceCompareTestsMixin):
    def setUp(self):
        self.recur = date.Recurrence('yearly')
        
    def testJanuary1(self):
        self.assertEqual(date.DateTime(2002,1,1), 
                         self.recur(date.DateTime(2001,1,1)))

    def testJanuary1_LeapYear(self):
        self.assertEqual(date.DateTime(2001,1,1), 
                         self.recur(date.DateTime(2000,1,1)))

    def testMarch1_LeapYear(self):
        self.assertEqual(date.DateTime(2001,3,1), 
                         self.recur(date.DateTime(2000,3,1)))
        
    def testMarch1_YearBeforeLeapYear(self):
        self.assertEqual(date.DateTime(2004,3,1), 
                         self.recur(date.DateTime(2003,3,1)))

    def testFebruary1_YearBeforeLeapYear(self):
        self.assertEqual(date.DateTime(2004,2,1), 
                         self.recur(date.DateTime(2003,2,1)))

    def testFebruary28(self):
        self.assertEqual(date.DateTime(2003,2,28), 
                         self.recur(date.DateTime(2002,2,28)))

    def testFebruary28_LeapYear(self):
        self.assertEqual(date.DateTime(2005,2,28), 
                         self.recur(date.DateTime(2004,2,28)))

    def testFebruary28_YearBeforeLeapYear(self):
        self.assertEqual(date.DateTime(2004,2,28), 
                         self.recur(date.DateTime(2003,2,28)))

    def testFebruary29(self):
        self.assertEqual(date.DateTime(2005,2,28), 
                         self.recur(date.DateTime(2004,2,29)))
                
        
class BiYearlyRecurrenceTest(test.TestCase, CommonRecurrenceTestsMixin,
                             YearlyRecurrenceCompareTestsMixin):
    def setUp(self):
        self.recur = date.Recurrence('yearly', amount=2)
        
    def testEveryOtherYear(self):
        self.assertEqual(date.DateTime(2004,3,1), 
                         self.recur(date.DateTime(2002,3,1)))
            

class YearlySameWeekDayRecurrenceTest(test.TestCase, CommonRecurrenceTestsMixin,
                                      YearlyRecurrenceCompareTestsMixin):
    def setUp(self):
        self.recur = date.Recurrence('yearly', sameWeekday=True)
        
    def testFirstTuesdayOfTheYear(self):
        self.assertEqual(date.DateTime(2009,1,6), 
                         self.recur(date.DateTime(2008,1,1)))

    def testFirstWednesdayOfTheYear(self):
        self.assertEqual(date.DateTime(2009,1,7), 
                         self.recur(date.DateTime(2008,1,2)))

    def testFirstThursdayOfTheYear(self):
        self.assertEqual(date.DateTime(2009,1,1), 
                         self.recur(date.DateTime(2008,1,3)))

    def testFirstFridayOfTheYear(self):
        self.assertEqual(date.DateTime(2009,1,2), 
                         self.recur(date.DateTime(2008,1,4)))

    def testLastWednesdayOfTheYear(self):
        self.assertEqual(date.DateTime(2009,12,30), 
                         self.recur(date.DateTime(2008,12,31)))

    def testLastTuesdayOfTheYear(self):
        self.assertEqual(date.DateTime(2009,12,29), 
                         self.recur(date.DateTime(2008,12,30)))

    def testLastMondayOfTheYear(self):
        self.assertEqual(date.DateTime(2009,12,28), 
                         self.recur(date.DateTime(2008,12,29)))

    def testLastSundayOfTheYear(self):
        self.assertEqual(date.DateTime(2009,12,27), 
                         self.recur(date.DateTime(2008,12,28)))

    def testLastSaturdayOfTheYear(self):
        self.assertEqual(date.DateTime(2009,12,26), 
                         self.recur(date.DateTime(2008,12,27)))

    def testLastFridayOfTheYear(self):
        self.assertEqual(date.DateTime(2009,12,25), 
                         self.recur(date.DateTime(2008,12,26)))

    def testLastThursdayOfTheYear(self):
        self.assertEqual(date.DateTime(2009,12,24), 
                         self.recur(date.DateTime(2008,12,25)))

        
class MaxRecurrenceTest(test.TestCase, CommonRecurrenceTestsMixin):
    def setUp(self):
        self.recur = date.Recurrence('daily', maximum=4)
        
    def testFirst(self):
        self.assertEqual(date.DateTime(2000,1,2), 
                         self.recur(date.DateTime(2000,1,1), next=True))
        
    def testCountAfterFirst(self):
        self.recur(date.DateTime(2000,1,1), next=True)
        self.assertEqual(1, self.recur.count)
        
    def testLast(self):
        self.recur.count = 4
        self.assertEqual(None, self.recur(date.DateTime(2000,1,1), next=True))




'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx
import TreeCtrlTest
from unittests import dummy
from taskcoachlib import widgets


class TreeListCtrlTestCase(TreeCtrlTest.TreeCtrlTestCase):

    onSelect = getItemTooltipText = None

    def setUp(self):
        super(TreeListCtrlTestCase, self).setUp()
        self._columns = self.createColumns()
        self.treeCtrl = widgets.TreeListCtrl(self.frame, self.columns(), 
            self.getItemTooltipText,
            self.onSelect, dummy.DummyUICommand(), dummy.DummyUICommand())
        imageList = wx.ImageList(16, 16)
        for bitmapName in ['led_blue_icon', 'folder_blue_icon']:
            imageList.Add(wx.ArtProvider_GetBitmap(bitmapName, wx.ART_MENU, 
                          (16,16)))
        self.treeCtrl.AssignImageList(imageList) # pylint: disable=E1101

    def createColumns(self):
        names = ['treeColumn'] + ['column%d'%index for index in range(1, 5)]
        return [widgets.Column(name, name, ('view', 'whatever'), None) for name in names]
        
    def columns(self):
        return self._columns

    
class TreeListCtrlTest(TreeListCtrlTestCase, TreeCtrlTest.CommonTestsMixin):
    pass


class TreeListCtrlColumnsTest(TreeListCtrlTestCase):
    def setUp(self):
        super(TreeListCtrlColumnsTest, self).setUp()
        self.children[None] = [TreeCtrlTest.DummyDomainObject('item')]
        self.treeCtrl.RefreshAllItems(1)
        self.visibleColumns = self.columns()[1:]
        
    def assertColumns(self):
        # pylint: disable=E1101
        self.assertEqual(len(self.visibleColumns)+1, self.treeCtrl.GetColumnCount())
        item = self.treeCtrl.GetFirstChild(self.treeCtrl.GetRootItem())[0]
        for columnIndex in range(1, len(self.visibleColumns)):
            self.assertEqual('item', self.treeCtrl.GetItemText(item, columnIndex))
    
    def showColumn(self, name, show=True):
        column = widgets.Column(name, name, ('view', 'whatever'), None)
        self.treeCtrl.showColumn(column, show)
        if show:
            index = self.columns()[1:].index(column)
            self.visibleColumns.insert(index, column)
        else:
            self.visibleColumns.remove(column)
    
    def testAllColumnsVisible(self):
        self.assertColumns()
        
    def testHideColumn(self):
        self.showColumn('column1', False)
        self.assertColumns()
        
    def testHideLastColumn(self):
        lastColumnHeader = 'column%d'%len(self.visibleColumns)
        self.showColumn(lastColumnHeader, False)
        self.assertColumns()
        
    def testShowColumn(self):
        self.showColumn('column2', False)
        self.showColumn('column2', True)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx


class CheckListBox(wx.CheckListBox):
    ''' The wx.CheckListBox does not support client data on all platforms, 
        so we do it ourselves. '''
    
    def __init__(self, *args, **kwargs):
        super(CheckListBox, self).__init__(*args, **kwargs)
        self.__clientData = dict()
    
    def Append(self, item, clientData=None):
        index = super(CheckListBox, self).Append(item)
        if clientData:
            self.__clientData[index] = clientData
        return index
            
    def Insert(self, item, position, clientData=None):
        ''' We don't need this at the moment. '''
            
    def GetClientData(self, index):
        return self.__clientData[index] if index in self.__clientData else None
    
    def Clear(self, *args, **kwargs):
        super(CheckListBox, self).Clear(*args, **kwargs)
        self.__clientData.clear()

    def Delete(self, *args, **kwargs):
        ''' We don't need this at the moment. '''
        raise NotImplementedError



#! /usr/bin/env python
import wx, wx.lib


TimeLineSelectionEvent, EVT_TIMELINE_SELECTED = wx.lib.newevent.NewEvent()
TimeLineActivationEvent, EVT_TIMELINE_ACTIVATED = wx.lib.newevent.NewEvent()


class HotMap(object):
    ''' Keep track of which node is where. '''
    
    def __init__(self, parent=None):
        self.parent = parent
        self.nodes = []
        self.rects = {}
        self.children = {}
        super(HotMap, self).__init__()
        
    def append(self, node, rect):
        self.nodes.append(node)
        self.rects[node] = rect
        self.children[node] = HotMap(node)
        
    def __getitem__(self, node):
        return self.children[node]
    
    def findNodeAtPosition(self, position, parent=None):
        ''' Retrieve the node at the given position. '''
        for node, rect in self.rects.items():
            if rect.Contains(position):
                return self[node].findNodeAtPosition(position, node)
        return parent
    
    def firstNode(self):
        return self.nodes[0] if self.nodes else None
    
    def lastNode(self, parent=None):
        if self.nodes:
            last = self.nodes[-1]
            return self[last].lastNode(last)
        else:
            return parent
    
    def findNode(self, target):
        if target in self.nodes:
            return self
        for node in self.nodes:
            result = self[node].findNode(target)
            if result:
                return result
        return None
    
    def nextChild(self, target):
        index = self.nodes.index(target)
        index = min(index+1, len(self.nodes)-1)
        return self.nodes[index]
    
    def previousChild(self, target):
        index = self.nodes.index(target)
        index = max(index-1, 0)
        return self.nodes[index]
    
    def firstChild(self, target):
        children = self[target].nodes
        if children:
            return children[0]
        else:
            return target        
        
           
class TimeLine(wx.Panel):
    def __init__(self, *args, **kwargs):
        self.model = kwargs.pop('model', [])
        self.padding = kwargs.pop('padding', 3)
        self.adapter = kwargs.pop('adapter', DefaultAdapter())
        self.selectedNode = None
        self.backgroundColour = wx.WHITE
        self._buffer = wx.EmptyBitmap(20, 20) # Have a default buffer ready
        self.DEFAULT_PEN = wx.Pen(wx.BLACK, 1, wx.SOLID)
        self.SELECTED_PEN = wx.Pen(wx.WHITE, 2, wx.SOLID)
        kwargs['style'] = wx.TAB_TRAVERSAL|wx.NO_BORDER|wx.FULL_REPAINT_ON_RESIZE|wx.WANTS_CHARS
        super(TimeLine, self).__init__(*args, **kwargs) 
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_SIZE, self.OnSize )
        self.Bind(wx.EVT_LEFT_UP, self.OnClickRelease)
        self.Bind(wx.EVT_LEFT_DCLICK, self.OnDoubleClick)
        self.Bind(wx.EVT_KEY_UP, self.OnKeyUp)
        self.OnSize(None)

    def SetBackgroundColour(self, colour):
        self.backgroundColour = colour

    def Refresh(self):
        self.UpdateDrawing()
    
    def OnPaint(self, event):
        dc = wx.BufferedPaintDC(self, self._buffer)

    def OnSize(self, event):
        # The buffer is initialized in here, so that the buffer is always
        # the same size as the Window.
        width, height = self.GetClientSizeTuple()
        if width <= 0 or height <= 0:
            return
        # Make new off-screen bitmap: this bitmap will always have the
        # current drawing in it, so it can be used to save the image to
        # a file, or whatever.
        self._buffer = wx.EmptyBitmap(width, height)
        self.UpdateDrawing()
        
    def OnClickRelease(self, event):
        event.Skip()
        self.SetFocus()
        point = event.GetPosition()
        node = self.hot_map.findNodeAtPosition(point)
        self.SetSelected(node, point)

    def OnDoubleClick(self, event):
        point = event.GetPosition()
        node = self.hot_map.findNodeAtPosition(point)
        if node:
            wx.PostEvent(self, TimeLineActivationEvent(node=node, point=point))
            
    def OnKeyUp(self, event):
        event.Skip()
        if not self.hot_map:
            return
        if event.KeyCode == wx.WXK_HOME:
            self.SetSelected(self.hot_map.firstNode())
            return
        elif event.KeyCode == wx.WXK_END:
            self.SetSelected(self.hot_map.lastNode())
            return
        if not self.selectedNode:
            return
        if event.KeyCode == wx.WXK_RETURN:
            wx.PostEvent(self, TimeLineActivationEvent(node=self.selectedNode))
            return
        hot_map = self.hot_map.findNode(self.selectedNode)
        if hot_map is None:
            newSelection = self.hot_map.firstNode()
        elif event.KeyCode == wx.WXK_DOWN:
            newSelection = hot_map.nextChild(self.selectedNode)
        elif event.KeyCode == wx.WXK_UP:
            newSelection = hot_map.previousChild(self.selectedNode)
        elif event.KeyCode == wx.WXK_RIGHT:
            newSelection = hot_map.firstChild(self.selectedNode)
        elif event.KeyCode == wx.WXK_LEFT and hot_map.parent:
            newSelection = hot_map.parent
        else:
            newSelection = self.selectedNode
        self.SetSelected(newSelection)
        
    def GetSelected(self):
        return self.selectedNode

    def SetSelected(self, node, point=None):
        ''' Set the given node selected in the timeline widget '''
        if node == self.selectedNode:
            return
        self.selectedNode = node 
        self.Refresh()
        if node:
            wx.PostEvent(self, TimeLineSelectionEvent(node=node, point=point))

    def UpdateDrawing(self):
        dc = wx.BufferedDC(wx.ClientDC(self), self._buffer)
        self.Draw(dc)
        
    def Draw(self, dc):
        ''' Draw the timeline on the device context. '''
        self.hot_map = HotMap()
        dc.BeginDrawing()
        brush = wx.Brush(self.backgroundColour)
        dc.SetBackground(brush)
        dc.Clear()
        dc.SetFont(self.FontForLabels(dc))
        if self.model:
            bounds = self.adapter.bounds(self.model)
            self.min_start = float(min(bounds))
            self.max_stop = float(max(bounds))
            if self.max_stop - self.min_start < 100:
                self.max_stop += 100 
            self.length = self.max_stop - self.min_start
            self.width, self.height = dc.GetSize()
            labelHeight = dc.GetTextExtent('ABC')[1] + 2 # Leave room for time labels
            self.DrawParallelChildren(dc, self.model, labelHeight, self.height-labelHeight, self.hot_map)
            self.DrawNow(dc)
        dc.EndDrawing()
        
    def FontForLabels(self, dc):
        ''' Return the default GUI font, scaled for printing if necessary. '''
        font = wx.SystemSettings_GetFont(wx.SYS_DEFAULT_GUI_FONT)
        scale = dc.GetPPI()[0] / wx.ScreenDC().GetPPI()[0]
        font.SetPointSize(scale*font.GetPointSize())
        return font
    
    def DrawBox(self, dc, node, y, h, hot_map, isSequentialNode=False, depth=0):
        if h < self.padding:
            return
        start, stop = self.adapter.start(node), self.adapter.stop(node)
        if start is None:
            start = self.min_start - 10
        if stop is None:
            stop = self.max_stop + 10
        start, stop = min(start, stop), max(start, stop) # Sanitize input
        x = self.scaleX(start) + 2*depth
        w = self.scaleWidth(stop - start) - 4*depth
        hot_map.append(node, (wx.Rect(int(x), int(y), int(w), int(h))))
        self.DrawRectangle(dc, node, x, y, w, h, isSequentialNode, depth)
        if not isSequentialNode:
            self.DrawIconAndLabel(dc, node, x, y, w, h, depth)
            seqHeight = min(dc.GetTextExtent('ABC')[1] + 2, h)     
            self.DrawSequentialChildren(dc, node, y+2, seqHeight-4, hot_map[node], depth+1)
            self.DrawParallelChildren(dc, node, y+seqHeight, h-seqHeight, hot_map[node], depth+1)
    
    def DrawRectangle(self, dc, node, x, y, w, h, isSequentialNode, depth):
        dc = wx.GCDC(dc) if isSequentialNode else dc
        dc.SetClippingRegion(x, y, w, h)
        dc.SetBrush(self.brushForNode(node, isSequentialNode, depth))
        dc.SetPen(self.penForNode(node, isSequentialNode, depth))
        rounding = 0 if isSequentialNode and (h < self.padding * 4 or w < self.padding * 4) else self.padding * 2
        dc.DrawRoundedRectangle(x, y, w, h, rounding)
        dc.DestroyClippingRegion()
        
    def DrawIconAndLabel(self, dc, node, x, y, w, h, depth):
        ''' Draw the icon, if any, and the label, if any, of the node. '''
        # Make sure the Icon and Label are visible:
        if x < 0:
            w -= abs(x)
            x = 0
        dc.SetClippingRegion(x+1, y+1, w-2, h-2) # Don't draw outside the box
        icon = self.adapter.icon(node, node==self.selectedNode)
        if icon and h >= icon.GetHeight() and w >= icon.GetWidth():
            iconWidth = icon.GetWidth() + 2
            dc.DrawIcon(icon, x+2, y+2) 
        else:
            iconWidth = 0
        if h >= dc.GetTextExtent('ABC')[1]:
            dc.SetFont(self.fontForNode(dc, node, depth))
            dc.SetTextForeground(self.textForegroundForNode(node, depth))
            dc.DrawText(self.adapter.label(node), x + iconWidth + 2, y+2)
        dc.DestroyClippingRegion()
 
    def DrawParallelChildren(self, dc, parent, y, h, hot_map, depth=0):
        children = self.adapter.parallel_children(parent)
        if not children:
            return
        childY = y
        h -= len(children) # vertical space between children
        recursiveChildrenList = [self.adapter.parallel_children(child, recursive=True) \
                                 for child in children]
        recursiveChildrenCounts = [len(recursiveChildren) for recursiveChildren in recursiveChildrenList]
        recursiveChildHeight = h / float(len(children) + sum(recursiveChildrenCounts))
        for child, numberOfRecursiveChildren in zip(children, recursiveChildrenCounts):
            childHeight = recursiveChildHeight * (numberOfRecursiveChildren + 1)
            if childHeight >= self.padding:
                self.DrawBox(dc, child, childY, childHeight, hot_map, depth=depth)
            childY += childHeight + 1

    def DrawSequentialChildren(self, dc, parent, y, h, hot_map, depth=0):
        for child in self.adapter.sequential_children(parent):
            self.DrawBox(dc, child, y, h, hot_map, isSequentialNode=True, depth=depth)
        
    def DrawNow(self, dc):
        alpha_dc = wx.GCDC(dc)
        alpha_dc.SetPen(wx.Pen(wx.Colour(128, 200, 128, 128), width=3))
        now = self.scaleX(self.adapter.now())
        alpha_dc.DrawLine(now, 0, now, self.height)
        label = self.adapter.nowlabel()
        textWidth = alpha_dc.GetTextExtent(label)[0]
        alpha_dc.DrawText(label, now - (textWidth / 2), 0)

    def scaleX(self, x):
        return self.scaleWidth(x - self.min_start)

    def scaleWidth(self, width):
        return (width / self.length) * self.width

    def textForegroundForNode(self, node, depth=0):
        ''' Determine the text foreground color to use to display the label of
            the given node '''
        if node == self.selectedNode:
            fg_color = wx.SystemSettings_GetColour(wx.SYS_COLOUR_HIGHLIGHTTEXT)
        else:
            fg_color = self.adapter.foreground_color(node, depth)
            if not fg_color:
                fg_color = wx.SystemSettings_GetColour(wx.SYS_COLOUR_WINDOWTEXT)
        return fg_color
    
    def fontForNode(self, dc, node, depth=0):
        ''' Determine the font to use to display the label of the given node,
            scaled for printing if necessary. '''
        font = self.adapter.font(node, depth)
        font = font if font else wx.SystemSettings_GetFont(wx.SYS_DEFAULT_GUI_FONT)
        scale = dc.GetPPI()[0] / wx.ScreenDC().GetPPI()[0]
        font.SetPointSize(scale*font.GetPointSize())
        return font

    def brushForNode(self, node, isSequentialNode=False, depth=0):
        ''' Create brush to use to display the given node '''
        if node == self.selectedNode:
            color = wx.SystemSettings_GetColour(wx.SYS_COLOUR_HIGHLIGHT)
        else:
            color = self.adapter.background_color(node)
            if color:
                # The adapter returns a 3-tuple
                color = wx.Colour(*color)
            else:
                red = (depth * 10)%255
                green = 255-((depth * 10)%255)
                blue = 200
                color = wx.Colour(red, green, blue)
        if isSequentialNode:
            color.Set(color.Red(), color.Green(), color.Blue(), 128)
        return wx.Brush(color)
    
    def penForNode(self, node, isSequentialNode=False, depth=0):
        ''' Determine the pen to use to display the given node '''
        pen = self.SELECTED_PEN if node == self.selectedNode else self.DEFAULT_PEN
        #style = wx.DOT if isSequentialNode else wx.SOLID 
        #pen.SetStyle(style)
        return pen


class DefaultAdapter(object):
    def parallel_children(self, node, recursive=False):
        children = node.parallel_children[:]
        if recursive:
            for child in node.parallel_children:
                children.extend(self.parallel_children(child, True))
        return children

    def sequential_children(self, node):
        return node.sequential_children

    def children(self, node):
        return self.parallel_children(node) + self.sequential_children(node)
    
    def bounds(self, node):
        times = [node.start, node.stop]
        for child in self.children(node):
            times.extend(self.bounds(child))
        return min(times), max(times)

    def start(self, node, recursive=False):
        starts = [node.start] 
        if recursive:
            starts.extend([self.start(child, True) \
                           for child in self.children(node)])
        return float(min(starts))

    def stop(self, node, recursive=False):
        stops = [node.stop]
        if recursive:
            stops.extend([self.stop(child, True) \
                          for child in self.children(node)])
        return float(max(stops))

    def label(self, node):
        return node.path
    
    def background_color(self, node):
        return None

    def foreground_color(self, node, depth):
        return None
    
    def icon(self, node):
        return None
    
    def now(self):
        return 0
    
    def nowlabel(self):
        return 'Now'
    
    
class TestApp(wx.App):
    ''' Basic application for holding the viewing Frame '''

    def __init__(self, size):
        self.size = size
        super(TestApp, self).__init__(0)

    def OnInit(self):
        ''' Initialise the application. '''
        wx.InitAllImageHandlers()
        self.frame = wx.Frame(None)
        self.frame.CreateStatusBar()
        model = self.get_model(self.size) 
        self.timeline = TimeLine(self.frame, model=model)
        self.frame.Show(True)
        return True

    def get_model(self, size):
        parallel_children, sequential_children = [], []
        if size > 0:
            parallel_children = [self.get_model(size-1) for i in range(size)]
        sequential_children = [Node('Seq 1', 30+10*size, 40+10*size, [], []),
                               Node('Seq 2', 80-10*size, 90-10*size, [], [])] 
        return Node('Node %d'%size, 0+5*size, 100-5*size, parallel_children, 
                    sequential_children)


class Node(object):
    def __init__(self, path, start, stop, subnodes, events):
        self.path = path
        self.start = start
        self.stop = stop
        self.parallel_children = subnodes 
        self.sequential_children = events

    def __repr__(self):
        return '%s(%r, %r, %r, %r, %r)'%(self.__class__.__name__, self.path, 
                                         self.start, self.stop, 
                                         self.parallel_children,
                                         self.sequential_children)
        

usage = 'timeline.py [size]'
        
def main():
    """Mainloop for the application"""
    import sys
    size = 3
    if len(sys.argv) > 1:
        if sys.argv[1] in ('-h', '--help'):
            print usage
        else:
            try:
                size = int(sys.argv[1])
            except ValueError:
                print usage
    else:
        app = TestApp(size)
        app.MainLoop()


if __name__ == "__main__":
    main()



# Test when we don't have a config file
# Since we don't know if the user will have a config file or not when
# running this test, we will move it out of the way and then move it back
# when we're done

import os
from gntp.test import GNTPTestCase
from gntp.config import GrowlNotifier

ORIGINAL_CONFIG = os.path.expanduser('~/.gntp')
BACKUP_CONFIG = os.path.expanduser('~/.gntp.backup')


class ConfigTests(GNTPTestCase):
	def setUp(self):
		if os.path.exists(ORIGINAL_CONFIG):
			os.rename(ORIGINAL_CONFIG, BACKUP_CONFIG)
		self.growl = GrowlNotifier(self.application, [self.notification_name])
		self.growl.register()

	def test_missing_config(self):
		self.assertIsTrue(self._notify(description='No config file test'))

	def tearDown(self):
		if os.path.exists(BACKUP_CONFIG):
			os.rename(BACKUP_CONFIG, ORIGINAL_CONFIG)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

'''
In place editors for viewers.
''' # pylint: disable=W0105

import wx
from taskcoachlib.thirdparty import hypertreelist
from taskcoachlib import widgets
from taskcoachlib.domain import date


class KillFocusAcceptsEditsMixin(object):
    ''' Mixin class to let in place editors accept changes whenever the user
        clicks outside the edit control instead of cancelling the changes. '''
    def StopEditing(self):
        try:
            if self.__has_focus():
                # User hit Escape
                super(KillFocusAcceptsEditsMixin, self).StopEditing()
            else:
                # User clicked outside edit window
                self.AcceptChanges()
                self.Finish()
        except wx.PyDeadObjectError:
            pass

    def __has_focus(self):
        ''' Return whether this control has the focus. '''

        def window_and_all_children(window):
            window_and_children = [window]
            for child in window.GetChildren():
                window_and_children.extend(window_and_all_children(child))
            return window_and_children

        return wx.Window.FindFocus() in window_and_all_children(self)


class SubjectCtrl(KillFocusAcceptsEditsMixin, hypertreelist.EditTextCtrl):
    ''' Single line inline control for editing item subjects. '''
    pass


class DescriptionCtrl(KillFocusAcceptsEditsMixin, hypertreelist.EditTextCtrl):
    ''' Multiline inline text control for editing item descriptions. '''
    def __init__(self, *args, **kwargs):
        kwargs['style'] = kwargs.get('style', 0) | wx.TE_MULTILINE
        super(DescriptionCtrl, self).__init__(*args, **kwargs)


class EscapeKeyMixin(object):
    ''' Mixin class for text(like) controls to properly handle the Escape key. 
        The inheriting class needs to bind to the event handler. For example:
        self._spinCtrl.Bind(wx.EVT_KEY_DOWN, self.OnKeyDown) '''
    def OnKeyDown(self, event):
        keyCode = event.GetKeyCode()
        if keyCode == wx.WXK_ESCAPE:
            self.StopEditing()
        elif keyCode in (wx.WXK_RETURN, wx.WXK_NUMPAD_ENTER) and not event.ShiftDown():
            # Notify the owner about the changes
            self.AcceptChanges()
            # Even if vetoed, close the control (consistent with MSW)
            wx.CallAfter(self.Finish)
        else:
            event.Skip()
      

class _SpinCtrl(EscapeKeyMixin, KillFocusAcceptsEditsMixin, 
                hypertreelist.EditCtrl, widgets.SpinCtrl):
    ''' Base spin control class. '''
    def __init__(self, parent, wxId, item, column, owner, value, 
                 *args, **kwargs):
        super(_SpinCtrl, self).__init__(parent, wxId, item, column, owner, 
                                        str(value), *args, **kwargs)
        self._textCtrl.Bind(wx.EVT_KEY_DOWN, self.OnKeyDown)

        
class PriorityCtrl(_SpinCtrl):
    ''' Spin control for priority. Since priorities can be any negative or
        positive integer we don't need to set an allowed range. '''
    pass

            
class PercentageCtrl(_SpinCtrl):
    ''' Spin control for percentages. '''
    def __init__(self, *args, **kwargs):
        super(PercentageCtrl, self).__init__(min=0, max=100, *args, **kwargs)
        

class Panel(wx.Panel):
    ''' Panel class for inline controls that need to be put into a panel. '''
    def __init__(self, parent, wxId, value, *args, **kwargs): # pylint: disable=W0613
        # Don't pass the value argument to the wx.Panel since it doesn't take 
        # a value argument
        super(Panel, self).__init__(parent, wxId, *args, **kwargs)      
        
    def makeSizer(self, control):  
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(control, flag=wx.EXPAND)
        self.SetSizerAndFit(sizer)


class BudgetCtrl(EscapeKeyMixin, KillFocusAcceptsEditsMixin, 
                 hypertreelist.EditCtrl, Panel):
    ''' Masked inline text control for editing budgets: 
        <hours>:<minutes>:<seconds>. '''
    def __init__(self, parent, wxId, item, column, owner, value):
        super(BudgetCtrl, self).__init__(parent, wxId, item, column, owner)
        hours, minutes, seconds = value.hoursMinutesSeconds()
        # Can't inherit from TimeDeltaCtrl because we need to override GetValue,
        # so we use composition instead
        self.__timeDeltaCtrl = widgets.masked.TimeDeltaCtrl(self, hours, 
                                                            minutes, seconds)
        self.__timeDeltaCtrl.Bind(wx.EVT_KEY_DOWN, self.OnKeyDown)
        self.makeSizer(self.__timeDeltaCtrl)
        
    def GetValue(self):
        return date.parseTimeDelta(self.__timeDeltaCtrl.GetValue())


class AmountCtrl(EscapeKeyMixin, KillFocusAcceptsEditsMixin, 
                 hypertreelist.EditCtrl, Panel):
    ''' Masked inline text control for editing amounts (floats >= 0). '''
    def __init__(self, parent, wxId, item, column, owner, value):
        super(AmountCtrl, self).__init__(parent, wxId, item, column, owner)
        self.__floatCtrl = widgets.masked.AmountCtrl(self, value)
        self.__floatCtrl.Bind(wx.EVT_KEY_DOWN, self.OnKeyDown)
        self.makeSizer(self.__floatCtrl)
        
    def GetValue(self):
        return self.__floatCtrl.GetValue()
 
    
class DateTimeCtrl(KillFocusAcceptsEditsMixin, hypertreelist.EditCtrl, Panel):
    ''' Inline date and time picker control. '''
    def __init__(self, parent, wxId, item, column, owner, value, **kwargs):
        relative = kwargs.pop('relative', False)
        if relative:
            start = kwargs.pop('startDateTime', date.Now())
        super(DateTimeCtrl, self).__init__(parent, wxId, item, column, owner)
        settings = kwargs['settings']
        starthour = settings.getint('view', 'efforthourstart')
        endhour = settings.getint('view', 'efforthourend')
        interval = settings.getint('view', 'effortminuteinterval')
        self._dateTimeCtrl = widgets.DateTimeCtrl(self, starthour=starthour,
                                                   endhour=endhour, 
                                                   interval=interval, showRelative=relative)
        self._dateTimeCtrl.SetValue(value)
        if relative:
            self._dateTimeCtrl.SetRelativeChoicesStart(start=None if start == date.DateTime() else start)
        self.makeSizer(self._dateTimeCtrl)
                
    def GetValue(self):
        return self._dateTimeCtrl.GetValue()



from ..py30compat import unittest

from keyring.backend import KeyringBackend
from keyring.backends import multi
import keyring.errors

class MultipartKeyringWrapperTestCase(unittest.TestCase):

    """Test the wrapper that breaks passwords into smaller chunks"""

    class MockKeyring(KeyringBackend):

        priority = 1

        def __init__(self):
            self.passwords = {}

        def get_password(self, service, username):
            return self.passwords.get(service+username)

        def set_password(self, service, username, password):
            self.passwords[service+username] = password

        def delete_password(self, service, username):
            try:
                del self.passwords[service+username]
            except KeyError:
                raise keyring.errors.PasswordDeleteError('not found')

    def testViablePassThru(self):
        kr = multi.MultipartKeyringWrapper(self.MockKeyring())
        self.assertTrue(kr.viable)

    def testMissingPassword(self):
        wrapped_kr = self.MockKeyring()
        kr = multi.MultipartKeyringWrapper(wrapped_kr)
        self.assertIsNone(kr.get_password('s1', 'u1'))

    def testSmallPasswordSetInSinglePart(self):
        wrapped_kr = self.MockKeyring()
        kr = multi.MultipartKeyringWrapper(wrapped_kr)
        kr.set_password('s1', 'u1', 'p1')
        self.assertEquals(wrapped_kr.passwords, {'s1u1':'p1'})
        # should be able to read it back
        self.assertEquals(kr.get_password('s1', 'u1'), 'p1')

    def testLargePasswordSetInMultipleParts(self):
        wrapped_kr = self.MockKeyring()
        kr = multi.MultipartKeyringWrapper(wrapped_kr,
            max_password_size=2)
        kr.set_password('s2', 'u2', '0123456')
        self.assertEquals(wrapped_kr.passwords, {'s2u2':'01',
                                                 's2u2{{part_1}}':'23',
                                                 's2u2{{part_2}}':'45',
                                                 "s2u2{{part_3}}":'6'})

        # should be able to read it back
        self.assertEquals(kr.get_password('s2', 'u2'), '0123456')
